# -*- coding: utf-8 -*-
"""
Tests for utilities related to python structures such as lists.
"""
import pytest

from nessai.utils.structures import replace_in_list


def test_replace_in_list():
    """
    Test if the list produced contains the correct entries in the correct
    locations
    """
    x = [1, 2, 3]
    replace_in_list(x, [1, 2], [5, 4])
    assert x == [5, 4, 3]


def test_replace_in_list_item():
    """
    Test if items are correctly converted to lists in replace_in_list function
    """
    x = [1, 2, 3]
    replace_in_list(x, 3, 4)
    assert x == [1, 2, 4]


def test_replace_in_list_single():
    """Test the function with a list of strings"""
    x = ['aa', 'bb', 'cc']
    replace_in_list(x, ['aa', 'bb'], ['dd', 'ee'])
    assert x == ['dd', 'ee', 'cc']


def test_replace_in_list_single_str():
    """Test the function with single strings"""
    x = ['aa', 'bb', 'cc']
    replace_in_list(x, 'aa', 'dd')
    assert x == ['dd', 'bb', 'cc']


def test_different_lengths():
    """
    Assert an error is raised if the targets and replacements are different
    lengths.
    """
    with pytest.raises(RuntimeError) as excinfo:
        replace_in_list([1, 2], [1, 2], 3)
    assert 'Targets and replacements are different lengths!' \
        in str(excinfo.value)


def test_missing_targets():
    """Assert an error is raised if a target is not in the target list."""
    with pytest.raises(ValueError) as excinfo:
        replace_in_list([1, 2], 4, 3)
    assert 'Targets [4] not in list: [1, 2]' in str(excinfo.value)



# -*- coding: utf-8 -*-
"""
Test the Rescale class.
"""
import numpy as np
import pytest
from unittest.mock import create_autospec

from nessai.livepoint import numpy_array_to_live_points
from nessai.reparameterisations import Rescale


@pytest.fixture()
def reparam():
    return create_autospec(Rescale)


@pytest.mark.parametrize('scale', [2, 2.0, [1, 2], {'x': 1, 'y': 2}])
def test_init(scale):
    """Test the init method with different input types"""
    parameters = ['x', 'y']
    prior_bounds = {'x': [-1, 1], 'y': [-1, 1]}

    reparam = \
        Rescale(parameters=parameters, scale=scale, prior_bounds=prior_bounds)

    assert not set(reparam.scale.keys()) - set(parameters)
    assert isinstance(reparam.scale['x'], float)


@pytest.mark.parametrize('n', [1, 2])
def test_reparameterise(reparam, n):
    """Test the reparameterise method"""
    reparam.parameters = ['x', 'y']
    reparam.prime_parameters = ['x_prime', 'y_prime']
    reparam.scale = {'x': -2.0, 'y': 4.0}
    x = numpy_array_to_live_points(np.ones((n, 2)), reparam.parameters)
    x_prime = numpy_array_to_live_points(
        np.zeros((n, 2)), reparam.prime_parameters)
    log_j = np.zeros(n)

    x_out, x_prime_out, log_j_out = \
        Rescale.reparameterise(reparam, x, x_prime, log_j)

    assert np.array_equal(x, x_out)
    assert np.array_equal(log_j_out, -np.log(8 * np.ones(n)))
    assert (x_prime_out['x_prime'] == -0.5).all()
    assert (x_prime_out['y_prime'] == 0.25).all()


@pytest.mark.parametrize('scale', [1e60, 1e-60])
def test_reparameterise_overflow(reparam, scale):
    """Test the reparameterise method with very small and large scales.

    Checks precision to 14 decimal places.
    """
    reparam.parameters = ['x']
    reparam.prime_parameters = ['x_prime']
    reparam.scale = {'x': scale}
    x_array = np.arange(100.0, dtype=float)
    x = numpy_array_to_live_points(scale * x_array[:, np.newaxis],
                                   reparam.parameters)
    x_prime = numpy_array_to_live_points(np.ones((x_array.size, 1)),
                                         reparam.prime_parameters)
    log_j = np.zeros(x.size)

    x_out, x_prime_out, log_j_out = \
        Rescale.reparameterise(reparam, x, x_prime, log_j)

    np.testing.assert_array_almost_equal(x_array, x_prime_out['x_prime'],
                                         decimal=14)
    assert (log_j == -np.log(scale)).all()


@pytest.mark.parametrize('n', [1, 2])
def test_inverse_reparameterise(reparam, n):
    """Test the inverse reparameterise method"""
    reparam.parameters = ['x', 'y']
    reparam.prime_parameters = ['x_prime', 'y_prime']
    reparam.scale = {'x': -2.0, 'y': 4.0}
    x = numpy_array_to_live_points(np.zeros((n, 2)), reparam.parameters)
    x_prime = numpy_array_to_live_points(
        np.ones((n, 2)), reparam.prime_parameters)
    x_prime['x_prime'] *= -1
    log_j = np.zeros(n)

    x_out, x_prime_out, log_j_out = \
        Rescale.inverse_reparameterise(reparam, x, x_prime, log_j)

    assert np.array_equal(x_prime, x_prime_out)
    assert np.array_equal(log_j_out, np.log(8 * np.ones(n)))
    assert (x_out['x'] == 2.0).all()
    assert (x_out['y'] == 4.0).all()


@pytest.mark.parametrize('scale', [1e60, 1e-60])
def test_inverse_reparameterise_overflow(reparam, scale):
    """Test the inverse_reparameterise method with very small and large scales.
    """
    reparam.parameters = ['x']
    reparam.prime_parameters = ['x_prime']
    reparam.scale = {'x': scale}
    x_array = np.arange(100.0, dtype=float)
    x = numpy_array_to_live_points(np.ones((x_array.size, 1)),
                                   reparam.parameters)
    x_prime = numpy_array_to_live_points(x_array[:, np.newaxis],
                                         reparam.prime_parameters)
    log_j = np.zeros(x.size)

    x_out, x_prime_out, log_j_out = \
        Rescale.inverse_reparameterise(reparam, x, x_prime, log_j)

    np.testing.assert_array_equal(x_array * scale, x_out['x'])
    assert (log_j == np.log(scale)).all()


def test_init_no_scale():
    """Make sure an error is raised if the scale is not given"""
    with pytest.raises(RuntimeError) as excinfo:
        Rescale(scale=None)
    assert 'Must specify a scale!' in str(excinfo.value)


@pytest.mark.parametrize('scale', [[1], [1, 2, 3]])
def test_init_incorrect_scale_list(scale):
    """Make sure an error is raised if the scale is the incorrect length"""
    parameters = ['x', 'y']
    prior_bounds = {'x': [-1, 1], 'y': [-1, 1]}

    with pytest.raises(RuntimeError) as excinfo:
        Rescale(parameters=parameters, scale=scale, prior_bounds=prior_bounds)

    assert 'different length' in str(excinfo.value)


@pytest.mark.parametrize('scale', [{'x': 1}, {'x': 1, 'y': 1, 'z': 1}])
def test_init_incorrect_scale_dict(scale):
    """Make sure an error is raised if the scale keys to not match the \
            parameters.
    """
    parameters = ['x', 'y']
    prior_bounds = {'x': [-1, 1], 'y': [-1, 1]}

    with pytest.raises(RuntimeError) as excinfo:
        Rescale(parameters=parameters, scale=scale, prior_bounds=prior_bounds)

    assert 'Mismatched parameters' in str(excinfo.value)


def test_init_incorrect_scale_type():
    """Make sure an error is raised if the scale is the incorrect type"""
    parameters = ['x', 'y']
    prior_bounds = {'x': [-1, 1], 'y': [-1, 1]}

    with pytest.raises(TypeError) as excinfo:
        Rescale(parameters=parameters, scale='1', prior_bounds=prior_bounds)

    assert 'Scale input must be' in str(excinfo.value)



# -*- coding: utf-8 -*-
"""
Test the analytic proposal method.
"""
import datetime
import numpy as np
import pytest
from unittest.mock import Mock, create_autospec, patch

from nessai.livepoint import numpy_array_to_live_points
from nessai.proposal import AnalyticProposal


@pytest.fixture
def proposal():
    return create_autospec(AnalyticProposal)


def test_init(proposal):
    """Test the init method."""
    with patch('nessai.proposal.analytic.Proposal.__init__') as mock_super:
        AnalyticProposal.__init__(proposal, 'model', poolsize=10, test=True)
    mock_super.assert_called_once_with('model', test=True)
    assert proposal._poolsize == 10
    assert proposal.populated is False


def test_poolsize(proposal):
    """Test poolsize property"""
    proposal._poolsize = 100
    poolsize = AnalyticProposal.poolsize.__get__(proposal)
    assert poolsize == 100


@pytest.mark.parametrize('pool', [None, True])
@pytest.mark.parametrize('N', [None, 5])
def test_populate(proposal, pool, N):
    """Test the populate process"""
    poolsize = 10
    if N is None:
        samples = numpy_array_to_live_points(
            np.arange(poolsize)[:, np.newaxis], 'x'
        )
        log_p = np.arange(poolsize, 2 * poolsize)
    else:
        samples = numpy_array_to_live_points(np.arange(N)[:, np.newaxis], 'x')
        log_p = np.arange(N, 2 * N)
    proposal.pool = pool
    proposal.poolsize = poolsize
    proposal.model = Mock()
    proposal.model.new_point = Mock(return_value=samples)
    proposal.model.log_prior = Mock(return_value=log_p)
    proposal.evaluate_likelihoods = Mock()
    AnalyticProposal.populate(proposal, N=N)

    if N is None:
        N = poolsize
    proposal.model.new_point.assert_called_once_with(N=N)
    proposal.model.log_prior.assert_called_once_with(samples)
    if pool is not None:
        proposal.evaluate_likelihoods.assert_called_once()
    else:
        proposal.evaluate_likelihoods.assert_not_called()

    np.testing.assert_array_equal(proposal.samples['logP'], log_p)
    assert sorted(proposal.indices) == list(range(N))
    assert proposal.populated is True


@pytest.mark.parametrize('populated', [True, False])
def test_draw(proposal, populated):
    """Test the draw method"""
    N = 5
    proposal.populated = populated
    proposal.populate = Mock()
    proposal.population_time = datetime.timedelta()
    proposal.indices = [0, 1, 2, 3, 4]
    proposal.samples = np.array([0, 1, 2, 3, 4])

    sample = AnalyticProposal.draw(proposal, 1, N=N)

    assert sample == 4

    if not populated:
        proposal.populate.assert_called_once_with(N=N)
        assert proposal.population_time.total_seconds() > 0.
    else:
        proposal.populate.assert_not_called()
        assert proposal.population_time.total_seconds() == 0.
    assert proposal.indices == [0, 1, 2, 3]


def test_draw_out_of_samples(proposal):
    """Assert populated is set to false when the last sample is used."""
    N = 5
    proposal.populated = True
    proposal.indices = [0]
    proposal.samples = np.array([0, 1, 2, 3, 4])

    sample = AnalyticProposal.draw(proposal, 1, N=N)

    assert sample == 0
    assert proposal.indices == []
    assert proposal.populated is False


@pytest.mark.integration_test
def test_draw_intergration(model):
    """Integration test for the draw method"""
    proposal = AnalyticProposal(model)
    N = 10
    old_point = model.new_point()
    assert not proposal.populated
    proposal.draw(old_point, N=N)
    assert proposal.populated
    [proposal.draw(old_point) for _ in range(N - 1)]
    assert not proposal.indices
    assert not proposal.populated


@pytest.mark.integration_test
def test_populate_integration(model):
    """Integration test for the populate method"""
    proposal = AnalyticProposal(model)
    N = 500
    proposal.populate(N=N)
    assert proposal.samples.size == N
    assert proposal.populated is True




from numpy.random import seed
import pytest
from scipy.stats import norm
import torch

from nessai.model import Model


seed(170817)
torch.manual_seed(170817)

_requires_dependency_cache = dict()


@pytest.fixture()
def model():
    class TestModel(Model):

        def __init__(self):
            self.bounds = {'x': [-5, 5], 'y': [-5, 5]}
            self.names = ['x', 'y']

        def log_prior(self, x):
            log_p = 0.
            for n in self.names:
                log_p += ((x[n] >= self.bounds[n][0])
                          & (x[n] <= self.bounds[n][1])) \
                        / (self.bounds[n][1] - self.bounds[n][0])
            return log_p

        def log_likelihood(self, x):
            log_l = 0
            for pn in self.names:
                log_l += norm.logpdf(x[pn])
            return log_l

    return TestModel()


@pytest.fixture()
def flow_config():
    d = dict(
            max_epochs=5,
            model_config=dict(n_blocks=2, n_neurons=2, n_layers=1,
                              device_tag='cpu',
                              kwargs=dict(batch_norm_between_layers=False))
            )
    return d


def pytest_configure(config):
    # register an additional marker
    config.addinivalue_line(
        "markers", "cuda: mark test to indicate it requires CUDA"
    )
    config.addinivalue_line(
        "markers",
        "requires(package): mark test to only run if the package can be "
        "imported"
    )


def pytest_runtest_setup(item):
    for mark in item.iter_markers(name='cuda'):
        if not torch.cuda.is_available():
            pytest.skip('Test requires CUDA')
    for mark in item.iter_markers(name='requires'):
        name = mark.args[0]
        if name in _requires_dependency_cache:
            skip_it = _requires_dependency_cache[name]
        else:
            try:
                __import__(name)
                skip_it = False
            except ImportError:
                skip_it = True

            _requires_dependency_cache[name] = skip_it

        reason = 'Missing dependency: {}'.format(name)
        if skip_it:
            pytest.skip(reason)



# -*- coding: utf-8 -*-
"""
Tests to ensure that the new GW proposal is equivalent to the old method
"""
import copy
import logging

import numpy as np

import pytest

from nessai.model import Model as BaseModel
from nessai.gw.proposal import GWFlowProposal
from nessai.gw.legacy import LegacyGWFlowProposal
from nessai.flowsampler import FlowSampler
from nessai.livepoint import dict_to_live_points

try:
    import bilby
except ImportError:
    print('Could not import bilby, some tests will not work')


logger = logging.getLogger(__name__)


@pytest.fixture(scope='module')
def injection_parameters():
    injection_parameters = dict(
        mass_1=36., mass_2=29., a_1=0.4, a_2=0.3, tilt_1=0.5, tilt_2=1.0,
        phi_12=1.7, phi_jl=0.3, luminosity_distance=2000., theta_jn=0.4,
        psi=2.659, phase=1.3, geocent_time=1126259642.413, ra=1.375,
        dec=-1.2108)
    return injection_parameters


@pytest.fixture(scope='function')
def legacy_kwargs():
    kwargs = {
        "boundary_inversion": True,
        "update_bounds": True,
        "rescale_parameters": True,
        "analytic_priors": True,
        "reparameterisations": {
            "rescale_angles": ["psi"],
            "inversion": True,
            "convert_to_angle": {"a_1": "split", "a_2": "duplicate"},
            "use_x_prime_prior": True,
            "uniform_distance_parameter": False,
            "spin_conversion": False
        },
        "detect_edges": True,
        "inversion_type": {"mass_ratio": "duplicate", "dc3": "duplicate"},
    }
    return kwargs


@pytest.fixture(scope='function')
def kwargs(legacy_kwargs):
    kwargs = copy.deepcopy(legacy_kwargs)
    kwargs['reparameterisations'] = {
        "mass_ratio": {"reparameterisation": "mass_ratio", "prior": "uniform"},
        "chirp_mass": {"reparameterisation": "mass", "prior": "uniform"},
        "luminosity_distance": {"reparameterisation": "distance",
                                "prior": "uniform-comoving-volume"},
        "geocent_time": {"reparameterisation": "time", "prior": "uniform",
                         "update_bounds": False},
        "sky-ra-dec": {"parameters": ["ra", "dec"], "prior": "isotropic"},
        "psi": {"reparameterisation": "angle-pi"},
        "theta_jn": {"reparameterisation": "angle-sine"},
        "tilt_1": {"reparameterisation": "angle-sine"},
        "tilt_2": {"reparameterisation": "angle-sine"},
        "phi_12": {"reparameterisation": "angle-2pi"},
        "phi_jl": {"reparameterisation": "angle-2pi"},
        "a_1": {"reparameterisation": "to-cartesian", "prior": "uniform"},
        "a_2": {"reparameterisation": "to-cartesian", "prior": "uniform",
                "mode": "duplicate"}
    }
    kwargs.pop('inversion_type')
    return kwargs


@pytest.mark.requires('bilby')
@pytest.mark.requires('lal')
@pytest.mark.requires('astropy')
@pytest.mark.parametrize('parameters',
                         ['psi',
                          'theta_jn',
                          'tilt_1',
                          'tilt_2',
                          'phi_12',
                          'phi_jl',
                          'a_1',
                          'a_2',
                          ['ra', 'dec'],
                          ['chirp_mass', 'geocent_time'],
                          ['chirp_mass', 'mass_ratio'],
                          ['chirp_mass', 'luminosity_distance'],
                          ])
@pytest.mark.parametrize('compute_radius', [False, True])
def test_parameter(parameters, injection_parameters, kwargs, legacy_kwargs,
                   compute_radius, tmpdir):
    """
    Test that the two proposal methods are equivalent for a set of parameters.

    Checks if the resulting x prime parameters are the same and their
    log-Jacobian determinants are also the same.
    """
    outdir = str(tmpdir.mkdir('test'))
    priors = bilby.gw.prior.BBHPriorDict()
    fixed_params = ["chirp_mass", "mass_ratio", "phi_12", "phi_jl", "a_1",
                    "a_2", "tilt_1", "tilt_2", "ra", "dec",
                    "luminosity_distance", "geocent_time", "theta_jn", "psi"]
    try:
        fixed_params.remove(parameters)
    except ValueError:
        for p in parameters:
            fixed_params.remove(p)
    priors['geocent_time'] = bilby.core.prior.Uniform(
            minimum=injection_parameters['geocent_time'] - 0.1,
            maximum=injection_parameters['geocent_time'] + 0.1,
            name='geocent_time', latex_label='$t_c$', unit='$s$')
    for key in fixed_params:
        if key in injection_parameters:
            priors[key] = injection_parameters[key]
        else:
            priors.pop(key)

    priors['mass_1'] = injection_parameters['mass_1']
    priors['mass_2'] = injection_parameters['mass_2']

    waveform_generator = bilby.gw.WaveformGenerator(
        duration=1, sampling_frequency=256,
        frequency_domain_source_model=bilby.gw.source.lal_binary_black_hole)

    likelihood = bilby.gw.GravitationalWaveTransient(
        interferometers=['H1'],
        waveform_generator=waveform_generator,
        priors=priors,
        phase_marginalization=True,
        distance_marginalization=False,
        time_marginalization=False, reference_frame='sky'
        )

    likelihood = bilby.core.likelihood.ZeroLikelihood(likelihood)

    def log_prior(theta):
        return priors.ln_prob(theta, axis=0)

    search_parameter_keys = []
    for key in priors:
        if isinstance(priors[key], bilby.core.prior.Prior) \
                and priors[key].is_fixed is False:
            search_parameter_keys.append(key)

    def log_likelihood(theta):
        params = {
            key: t for key, t in zip(search_parameter_keys, theta)}

        likelihood.parameters.update(params)
        return likelihood.log_likelihood_ratio()

    class Model(BaseModel):
        def __init__(self, names, priors):
            self.names = names
            self.priors = priors
            self._update_bounds()

        @staticmethod
        def log_likelihood(x, **kwargs):
            theta = [x[n].item() for n in search_parameter_keys]
            return log_likelihood(theta)

        @staticmethod
        def log_prior(x, names=None, **kwargs):
            if names is None:
                names = search_parameter_keys
            theta = {n: x[n] for n in names}
            return log_prior(theta)

        def _update_bounds(self):
            self.bounds = \
                {key: [self.priors[key].minimum, self.priors[key].maximum]
                 for key in self.names}

        def new_point(self, N=1):
            prior_samples = self.priors.sample(size=N)
            samples = {n: prior_samples[n] for n in self.names}
            self._update_bounds()
            return dict_to_live_points(samples)

        def new_point_log_prob(self, x):
            return self.log_prior(x)

    reparameterisations = kwargs.pop('reparameterisations')
    for k in list(reparameterisations.keys()):
        if k == 'sky-ra-dec':
            if any(p in fixed_params for p in ['ra', 'dec']):
                del reparameterisations['sky-ra-dec']
        elif k not in search_parameter_keys:
            del reparameterisations[k]

    if 'luminosity_distance' in parameters:
        legacy_kwargs['reparameterisations']['uniform_distance_parameter'] = \
                True

    model = Model(search_parameter_keys, priors)
    sampler = FlowSampler(model, output=outdir,
                          flow_class=LegacyGWFlowProposal, **legacy_kwargs)

    new_sampler = FlowSampler(model, output=outdir,
                              reparameterisations=reparameterisations,
                              flow_class=GWFlowProposal, **kwargs)

    sampler.ns._flow_proposal.initialise()
    new_sampler.ns._flow_proposal.initialise()

    sampler.ns._uninformed_proposal.populate(N=100)
    x = sampler.ns._uninformed_proposal.samples

    orig_proposal = sampler.ns._flow_proposal
    new_proposal = new_sampler.ns._flow_proposal

    tests = ['lower', 'upper', False]
    test_results = {t: False for t in tests}
    for test in tests:
        orig_proposal.check_state(x)
        np.random.seed(1234)
        x_prime, lj = orig_proposal.rescale(
            x, test=test, compute_radius=compute_radius)
        log_p = orig_proposal.x_prime_log_prior(x_prime)

        x_re, lj_re = orig_proposal.inverse_rescale(x_prime)

        new_proposal.check_state(x)
        np.random.seed(1234)
        x_prime_new, lj_new = new_proposal.rescale(
            x, test=test, compute_radius=compute_radius)
        log_p_new = new_proposal.x_prime_log_prior(x_prime_new)

        x_re_new, lj_re_new = new_proposal.inverse_rescale(x_prime_new)

        for n in model.names:
            np.testing.assert_array_almost_equal(x_re[n], x_re_new[n])

        np.testing.assert_array_equal(lj, lj_new)
        np.testing.assert_array_equal(lj_re, lj_re_new)
        np.testing.assert_array_equal(log_p, log_p_new)
        try:
            np.testing.assert_array_equal(x_prime, x_prime_new)
        except AssertionError:
            flag = {n: False for n in x_prime.dtype.names}
            for n in x_prime.dtype.names:
                for nn in x_prime_new.dtype.names:
                    if np.allclose(x_prime[n], x_prime_new[nn]):
                        logger.critical(f'{n} is equivalent to {nn}')
                        flag[n] = True

            assert all(v for v in flag.values()), print(f'Flags: {flag}')
        test_results[test] = True

    assert all(t for t in test_results.values())

    if getattr(orig_proposal, '_rescaled_min'):
        logger.info('Checking prime prior bounds')
        test_results = {t: False for t in tests}
        for test in tests:
            logger.info(f'Testing with inversion and test={test}')
            orig_proposal.check_state(x)
            np.random.seed(1234)
            t, lj = orig_proposal.rescale(x, test=test)
            new_proposal.check_state(x)
            np.random.seed(1234)
            t_new, lj_new = new_proposal.rescale(x, test=test)
            min_vals = \
                {n: False for n in new_proposal._reparameterisation.keys()}
            max_vals = \
                {n: False for n in new_proposal._reparameterisation.keys()}
            for k, r in new_proposal._reparameterisation.items():
                for pb in r.prime_prior_bounds.values():
                    for vmin in orig_proposal._rescaled_min.values():
                        logger.warning([k, pb, vmin])
                        if np.isclose(pb[0], vmin):
                            min_vals[k] = True
                    for vmax in orig_proposal._rescaled_max.values():
                        logger.warning([k, pb, vmax])
                        if np.isclose(pb[1], vmax):
                            max_vals[k] = True

            assert all(t for t in min_vals.values()), print(min_vals)
            assert all(t for t in max_vals.values()), print(max_vals)
            test_results[test] = True

        assert all(t for t in test_results.values())



# -*- coding: utf-8 -*-
"""
Test utilities related to logging.
"""
import logging
import os
import pytest

from nessai.utils.logging import setup_logger


def teardown_function():
    """Reset the logger after each test"""
    logger = logging.getLogger('nessai')
    logger.handlers = []
    logger.addHandler(logging.NullHandler())
    try:
        os.remove('test.log')
    except OSError:
        pass


def test_setup_logger_no_label():
    """Test behaviour when label is None.

    This should NOT produce a log file.
    """
    logger = setup_logger(label=None)
    assert not any([type(h) == logging.FileHandler for h in logger.handlers])


@pytest.mark.parametrize('output', ['logger_dir', None])
def test_setup_logger_with_label(tmpdir, output):
    """Test behaviour when label is not None.

    This should produce a log file.
    """
    if output:
        output = str(tmpdir.mkdir(output))
    logger = setup_logger(label='test', output=output)
    if output is None:
        output = '.'
    assert os.path.exists(f'{output}/test.log')
    assert any([type(h) == logging.FileHandler for h in logger.handlers])


def test_setup_logger_with_mkdir(tmpdir):
    """Assert the output directory is created if missing"""
    output = str(tmpdir) + '/logger_dir/'
    setup_logger(label='test', output=output)
    assert os.path.exists(f'{output}/test.log')


@pytest.mark.parametrize(
    'log_level, value',
    [('ERROR', 40), ('WARNING', 30), ('INFO', 20), ('DEBUG', 10), (15, 15)]
)
def test_setup_logger_levels(log_level, value):
    """Verify logging levels are correctly set."""
    logger = setup_logger(log_level=log_level, label=None)
    assert all([h.level == value for h in logger.handlers])


def test_setup_logger_unknown_level():
    """Verify an error is raised for an unknown level"""
    with pytest.raises(ValueError) as excinfo:
        setup_logger(log_level='test', label=None)
    assert 'log_level test not understood' in str(excinfo.value)



# -*- coding: utf-8 -*-
"""
Integration tests for adding the default reparameterisations
"""
import numpy as np
import pytest
from unittest.mock import create_autospec

from nessai.model import Model
from nessai.proposal import FlowProposal
from nessai.reparameterisations import default_reparameterisations

# General reparameterisations that do not need extra parameters
general_reparameterisations = \
    {k: v for k, v in default_reparameterisations.items()
        if k not in ['scale', 'rescale', 'angle-pair']}


@pytest.fixture(params=general_reparameterisations.keys())
def reparameterisation(request):
    return request.param


@pytest.fixture
def model():
    m = create_autospec(Model)
    m.names = ['x']
    m.bounds = {'x': [-1, 1]}
    m.reparameterisations = None
    return m


@pytest.mark.integration_test
def test_configure_reparameterisations(tmpdir, model, reparameterisation):
    """Test adding one of the default reparameterisations.

    Only tests reparameterisations that don't need extra parameters.
    """
    proposal = FlowProposal(
        model,
        output=str(tmpdir.mkdir('test')),
        poolsize=10,
        reparameterisations={'x': reparameterisation}
        )
    proposal.set_rescaling()
    assert proposal._reparameterisation is not None


@pytest.mark.integration_test
@pytest.mark.parametrize('reparameterisation', ['scale', 'rescale'])
def test_configure_reparameterisation_scale(tmpdir, model, reparameterisation):
    """Test adding the `Rescale` reparameterisation"""
    proposal = FlowProposal(
        model,
        output=str(tmpdir.mkdir('test')),
        poolsize=10,
        reparameterisations={
            'x': {'reparameterisation': reparameterisation, 'scale': 2.0}
        }
        )
    proposal.set_rescaling()
    assert proposal._reparameterisation is not None


@pytest.mark.integration_test
def test_configure_reparameterisation_angle_pair(tmpdir, model):
    """Test adding the `AnglePair` reparameterisation"""
    model.names.append('y')
    model.bounds = {'x': [0, 2 * np.pi], 'y': [-np.pi / 2, np.pi / 2]}
    proposal = FlowProposal(
        model,
        output=str(tmpdir.mkdir('test')),
        poolsize=10,
        reparameterisations={
            'x': {'reparameterisation': 'angle-pair', 'parameters': ['y']}
        }
        )
    proposal.set_rescaling()
    assert proposal._reparameterisation is not None



# -*- coding: utf-8 -*-
"""
Test functions related to handling live points
"""
import numpy as np
import pytest
from unittest.mock import MagicMock, patch

from nessai.nestedsampler import NestedSampler


def test_insert_live_point(sampler):
    """Test inserting a live point"""
    sampler.live_points = np.arange(-5, 0, 1.0).view([('logL', 'f8')])
    new_point = np.array(-3.5, dtype=[('logL', 'f8')])
    index = NestedSampler.insert_live_point(sampler, new_point)
    assert index == 1


def test_populate_live_points(sampler):
    """Test popluting the live points"""
    sampler.yield_sample = MagicMock(
        return_value=iter(zip(np.ones(sampler.nlive),
                              sampler.model.new_point(sampler.nlive)))
        )
    NestedSampler.populate_live_points(sampler)
    assert len(sampler.live_points) == sampler.nlive


def test_populate_live_points_nans(sampler):
    """Test popluting the live points with NaN values"""
    new_points = sampler.model.new_point(sampler.nlive + 1)
    new_points['logL'][4] = np.nan
    sampler.yield_sample = MagicMock(
        return_value=iter(zip(np.ones(sampler.nlive + 1), new_points))
        )
    NestedSampler.populate_live_points(sampler)
    assert len(sampler.live_points) == sampler.nlive
    assert not np.isnan(sampler.live_points['logL']).any()


@pytest.mark.parametrize('rolling', [False, True])
@patch('nessai.nestedsampler.compute_indices_ks_test', return_value=(0.1, 0.5))
def test_insertion_indices(mock_fn, rolling, sampler):
    """Test computing the distribution of insertion indices"""
    sampler.rolling_p = []
    sampler.insertion_indices = \
        np.random.randint(sampler.nlive, size=2 * sampler.nlive)

    NestedSampler.check_insertion_indices(sampler, rolling=rolling)

    if rolling:
        assert len(sampler.rolling_p) == 1
        np.testing.assert_array_equal(
            mock_fn.call_args_list[0][0][0],
            sampler.insertion_indices[-sampler.nlive:])
    else:
        mock_fn.assert_called_once_with(
            sampler.insertion_indices, sampler.nlive)


@patch('nessai.nestedsampler.compute_indices_ks_test', return_value=(0, None))
def test_insertion_indices_p_none(mock_fn, sampler):
    """Test computing the distribution of insertion indices if p is None"""
    sampler.rolling_p = []
    sampler.insertion_indices = \
        np.random.randint(sampler.nlive, size=2 * sampler.nlive)

    NestedSampler.check_insertion_indices(sampler, rolling=True)

    assert len(sampler.rolling_p) == 0


@pytest.mark.parametrize('filename', [None, 'file.txt'])
@patch('numpy.savetxt')
@patch('nessai.nestedsampler.compute_indices_ks_test', return_value=(0.1, 0.5))
def test_insertion_indices_save(mock_fn, mock_save, filename, sampler):
    """Test saving the insertion indices"""
    sampler.output = './'
    sampler.insertion_indices = \
        np.random.randint(sampler.nlive, size=2 * sampler.nlive)

    NestedSampler.check_insertion_indices(sampler, rolling=False,
                                          filename=filename)

    if filename:
        mock_save.assert_called_once_with(
            './file.txt', sampler.insertion_indices, newline='\n',
            delimiter=' ')



#!/usr/bin/env python

"""
Example of running nessai with bilby on a gravitational wave likelihood using
the legacy implementation of the proposal method as used in the paper. This
method is less flexible and has hard coded settings which make it impractical
to build upon.

Based on the Bilby example: https://git.ligo.org/lscsoft/bilby
"""
import bilby
import numpy as np

outdir = './outdir/'
label = 'legacy_gw_example'

bilby.core.utils.setup_logger(outdir=outdir, label=label, log_level='INFO')

duration = 4.
sampling_frequency = 2048.

np.random.seed(151226)

# Use an injection that is similar to GW150914
injection_parameters = dict(
    total_mass=66., mass_ratio=0.9, a_1=0.4, a_2=0.3, tilt_1=0.5, tilt_2=1.0,
    phi_12=1.7, phi_jl=0.3, luminosity_distance=2000, theta_jn=0.4, psi=2.659,
    phase=1.3, geocent_time=1126259642.413, ra=1.375, dec=-1.2108)

waveform_arguments = dict(waveform_approximant='IMRPhenomPv2',
                          reference_frequency=50.)

# Create the waveform_generator
waveform_generator = bilby.gw.waveform_generator.WaveformGenerator(
    sampling_frequency=sampling_frequency,
    duration=duration,
    frequency_domain_source_model=bilby.gw.source.lal_binary_black_hole,
    parameter_conversion=(bilby.gw.conversion
                          .convert_to_lal_binary_black_hole_parameters),
    waveform_arguments=waveform_arguments)

# Set up interferometers
ifos = bilby.gw.detector.InterferometerList(['H1', 'L1', 'V1'])
ifos.set_strain_data_from_power_spectral_densities(
    sampling_frequency=sampling_frequency, duration=duration,
    start_time=injection_parameters['geocent_time'] - 3)
ifos.inject_signal(waveform_generator=waveform_generator,
                   parameters=injection_parameters)

# Set up prior
# Nessai is designed to sample mass ratio and chirp mass
priors = bilby.gw.prior.BBHPriorDict()
priors['chirp_mass'] = bilby.prior.Uniform(
    name='chirp_mass', latex_label='$m_c$', minimum=13, maximum=45,
    unit='$M_{\\odot}$')
priors['mass_ratio'] = bilby.prior.Uniform(
    name='mass_ratio', latex_label='q', minimum=0.125, maximum=1.0)

priors['geocent_time'] = bilby.core.prior.Uniform(
    minimum=injection_parameters['geocent_time'] - 0.1,
    maximum=injection_parameters['geocent_time'] + 0.1,
    name='geocent_time', latex_label='$t_c$', unit='$s$')

# Only sample a subset of the parameters
for key in ['a_1', 'a_2', 'tilt_1', 'tilt_2', 'phi_12', 'phi_jl',
            'luminosity_distance', 'psi', 'geocent_time', 'ra', 'dec']:
    priors[key] = injection_parameters[key]

# Initialise the likelihood
likelihood = bilby.gw.likelihood.GravitationalWaveTransient(
    interferometers=ifos, waveform_generator=waveform_generator,
    phase_marginalization=True, priors=priors)

# Configuration for the normalising flow
flow_config = {
    "lr": 0.001,
    "batch_size": 1000,
    "val_size": 0.1,
    "max_epochs": 200,
    "patience": 50,
    "model_config": {
        "n_blocks": 2,
        "n_neurons": 4,
        "n_layers": 2,
        "ftype": "realnvp",
        "kwargs": {
            "batch_norm_between_layers": True,
            "linear_transform": "lu"
        }
    }
}

# Run sampler

# The `flow_class` should be set to `GWFlowProposal` for GW PE. This includes
# specific default reparameterisations for certain parameters. For example,
# it knows that theta_jn is angle with a sine prior.

# This examples use the legacy GWFlowProposal class which was used in the
# paper but has a number of settings hard coded and is generally less flexible.
result = bilby.core.sampler.run_sampler(
    likelihood=likelihood,
    priors=priors,
    outdir=outdir,
    injection_parameters=injection_parameters,
    label=label,
    conversion_function=bilby.gw.conversion.generate_all_bbh_parameters,
    flow_class='LegacyGWFlowProposal',
    sampler='nessai',
    resume=False,
    plot=True,
    nlive=2000,
    maximum_uninformed=4000,
    flow_config=flow_config,
    seed=150914,
    analytic_priors=True,      # Bilby priors can be sampled from directly
    inversion_type={'mass_ratio': 'duplicate'}
    )

# Produce corner plots
result.plot_corner()



# -*- coding: utf-8 -*-
"""Test methods related to reparameterisations"""
import numpy as np
from nessai.livepoint import numpy_array_to_live_points
from nessai.proposal import FlowProposal
from nessai.reparameterisations import (
    get_reparameterisation
)
import pytest
from unittest.mock import MagicMock, Mock, call, patch


@pytest.fixture
def dummy_rc():
    """Dummy reparameteristation class"""
    m = MagicMock()
    m.__name__ = 'DummyReparameterisation'
    return m


@pytest.fixture
def dummy_cmb_rc():
    """Dummy combined reparameteristation class"""
    m = MagicMock()
    m.add_reparameterisation = MagicMock()
    return m


def test_default_reparameterisation(proposal):
    """Test to make sure default reparameterisation does not cause errors
    for default proposal.
    """
    FlowProposal.add_default_reparameterisations(proposal)


@patch('nessai.reparameterisations.get_reparameterisation')
def test_get_reparamaterisation(mocked_fn, proposal):
    """Make sure the underlying function is called"""
    FlowProposal.get_reparameterisation(proposal, 'angle')
    assert mocked_fn.called_once_with('angle')


def test_configure_reparameterisations_dict(proposal, dummy_cmb_rc, dummy_rc):
    """Test configuration for reparameterisations dictionary.

    Also tests to make sure boundary inversion is set.
    """
    dummy_rc.return_value = 'r'
    # Need to add the parameters before hand to prevent a
    # NullReparameterisation from being added
    dummy_cmb_rc.parameters = ['x']
    proposal.add_default_reparameterisations = MagicMock()
    proposal.get_reparameterisation = MagicMock(return_value=(
        dummy_rc, {'boundary_inversion': True}
    ))
    proposal.model = MagicMock()
    proposal.model.bounds = {'x': [-1, 1], 'y': [-1, 1]}
    proposal.names = ['x']

    with patch('nessai.proposal.flowproposal.CombinedReparameterisation',
               return_value=dummy_cmb_rc) as mocked_class:
        FlowProposal.configure_reparameterisations(
            proposal, {'x': {'reparameterisation': 'default'}})

    proposal.get_reparameterisation.assert_called_once_with('default')
    proposal.add_default_reparameterisations.assert_called_once()
    dummy_rc.assert_called_once_with(
        prior_bounds={'x': [-1, 1]},
        parameters='x',
        boundary_inversion=True
    )
    mocked_class.assert_called_once()
    proposal._reparameterisation.add_reparameterisations.\
        assert_called_once_with('r')

    assert proposal.boundary_inversion is True
    assert proposal.names == ['x']


@patch('nessai.proposal.flowproposal.CombinedReparameterisation')
def test_configure_reparameterisations_dict_w_params(mocked_class, proposal,
                                                     dummy_rc, dummy_cmb_rc):
    """Test configuration for reparameterisations dictionary with parameters.

    For example:

        {'x': {'reparmeterisation': 'default', 'parameters': 'y'}}

    This should add both x and y to the reparameterisation.
    """
    dummy_rc.return_value = 'r'
    # Need to add the parameters before hand to prevent a
    # NullReparameterisation from being added
    dummy_cmb_rc.parameters = ['x', 'y']
    proposal.add_default_reparameterisations = MagicMock()
    proposal.get_reparameterisation = MagicMock(return_value=(
        dummy_rc, {},
    ))
    proposal.model = MagicMock()
    proposal.model.bounds = {'x': [-1, 1], 'y': [-1, 1]}
    proposal.names = ['x', 'y']

    with patch('nessai.proposal.flowproposal.CombinedReparameterisation',
               return_value=dummy_cmb_rc) as mocked_class:
        FlowProposal.configure_reparameterisations(
            proposal,
            {'x': {'reparameterisation': 'default', 'parameters': ['y']}}
        )

    proposal.get_reparameterisation.assert_called_once_with('default')
    proposal.add_default_reparameterisations.assert_called_once()
    dummy_rc.assert_called_once_with(
        prior_bounds={'x': [-1, 1], 'y': [-1, 1]},
        parameters=['y', 'x'],
    )
    mocked_class.assert_called_once()
    proposal._reparameterisation.add_reparameterisations.\
        assert_called_once_with('r')

    assert proposal.names == ['x', 'y']


def test_configure_reparameterisations_requires_prime_prior(
        proposal,
        dummy_rc,
        dummy_cmb_rc
):
    """
    Test configuration that requires a prime prior but the prime prior is
    missing.
    """
    dummy_rc.return_value = 'r'
    # Need to add the parameters before hand to prevent a
    # NullReparameterisation from being added
    dummy_cmb_rc.parameters = ['x', 'y']
    dummy_cmb_rc.has_prime_prior = False
    dummy_cmb_rc.requires_prime_prior = True
    proposal.add_default_reparameterisations = MagicMock()
    proposal.get_reparameterisation = MagicMock(return_value=(
        dummy_rc, {},
    ))
    proposal.model = MagicMock()
    proposal.model.bounds = {'x': [-1, 1], 'y': [-1, 1]}
    proposal.names = ['x', 'y']

    with patch('nessai.proposal.flowproposal.CombinedReparameterisation',
               return_value=dummy_cmb_rc), \
         pytest.raises(RuntimeError) as excinfo:
        FlowProposal.configure_reparameterisations(
            proposal,
            {'x': {'reparameterisation': 'default', 'parameters': ['y']}}
        )

    assert 'One or more reparameterisations require ' in str(excinfo.value)


@patch('nessai.reparameterisations.CombinedReparameterisation')
def test_configure_reparameterisations_dict_missing(mocked_class, proposal):
    """
    Test configuration for reparameterisations dictionary when missing
    the reparameterisation for a parameter.

    This should raise a runtime error.
    """
    proposal.add_default_reparameterisations = MagicMock()
    proposal.get_reparameterisation = get_reparameterisation
    proposal.model = MagicMock
    proposal.model.bounds = {'x': [-1, 1], 'y': [-1, 1]}
    proposal.names = ['x', 'y']

    with pytest.raises(RuntimeError) as excinfo:
        FlowProposal.configure_reparameterisations(
            proposal, {'x': {'scale': 1.0}})

    assert 'No reparameterisation found for x' in str(excinfo.value)


@patch('nessai.reparameterisations.CombinedReparameterisation')
def test_configure_reparameterisations_str(mocked_class, proposal):
    """Test configuration for reparameterisations dictionary from a str"""
    proposal.add_default_reparameterisations = MagicMock()
    proposal.get_reparameterisation = get_reparameterisation
    proposal.model = MagicMock
    proposal.model.bounds = {'x': [-1, 1], 'y': [-1, 1]}
    proposal.names = ['x', 'y']
    FlowProposal.configure_reparameterisations(
        proposal, {'x': 'default'})

    proposal.add_default_reparameterisations.assert_called_once()
    assert proposal.rescaled_names == ['x_prime', 'y']
    assert proposal.rescale_parameters == ['x']
    assert proposal._reparameterisation.parameters == ['x', 'y']
    assert proposal._reparameterisation.prime_parameters == ['x_prime', 'y']
    assert mocked_class.called_once


@patch('nessai.reparameterisations.CombinedReparameterisation')
def test_configure_reparameterisations_dict_reparam(mocked_class, proposal):
    """Test configuration for reparameterisations dictionary"""
    proposal.add_default_reparameterisations = MagicMock()
    proposal.get_reparameterisation = get_reparameterisation
    proposal.model = MagicMock
    proposal.model.bounds = {'x': [-1, 1], 'y': [-1, 1]}
    proposal.names = ['x', 'y']
    FlowProposal.configure_reparameterisations(
        proposal, {'default': {'parameters': ['x']}})

    proposal.add_default_reparameterisations.assert_called_once()
    assert proposal.rescaled_names == ['x_prime', 'y']
    assert proposal.rescale_parameters == ['x']
    assert proposal._reparameterisation.parameters == ['x', 'y']
    assert proposal._reparameterisation.prime_parameters == ['x_prime', 'y']
    assert mocked_class.called_once


def test_configure_reparameterisations_incorrect_type(proposal):
    """Assert an error is raised when input is not a dictionary"""
    with pytest.raises(TypeError) as excinfo:
        FlowProposal.configure_reparameterisations(proposal, ['default'])
    assert 'must be a dictionary' in str(excinfo.value)


def test_configure_reparameterisations_incorrect_config_type(proposal):
    """Assert an error is raised when the config for a key is not a dictionary
    or a known reparameterisation.
    """
    proposal.names = ['x']
    with pytest.raises(TypeError) as excinfo:
        FlowProposal.configure_reparameterisations(proposal, {'x': ['a']})
    assert 'Unknown config type' in str(excinfo.value)


@pytest.mark.parametrize(
    'reparam',
    [{'z': {'reparameterisation': 'sine'}}, {'sine': {'parameters': ['x']}}]
)
def test_configure_reparameterisation_unknown(proposal, reparam):
    """
    Assert an error is raised if an unknown reparameterisation or parameters
    is passed.
    """
    proposal.names = ['x']
    with pytest.raises(RuntimeError) as excinfo:
        FlowProposal.configure_reparameterisations(proposal, reparam)
    assert 'is not a parameter in the model or a known' in str(excinfo.value)


def test_configure_reparameterisation_no_parameters(proposal, dummy_rc):
    """Assert an error is raised if no parameters are specified"""
    proposal.names = ['x']
    proposal.get_reparameterisation = MagicMock(return_value=(
        dummy_rc, {},
    ))
    with pytest.raises(RuntimeError) as excinfo:
        FlowProposal.configure_reparameterisations(
            proposal, {'default': {'update_bounds': True}})
    assert 'No parameters key' in str(excinfo.value)


def test_set_rescaling_with_model(proposal, model):
    """
    Test setting the rescaling when the model contains reparmaeterisations.
    """
    proposal.model = model
    proposal.model.reparameterisations = {'x': 'default'}
    proposal.set_boundary_inversion = MagicMock()

    def update(self):
        proposal.rescale_parameters = ['x']
        proposal.rescaled_names = ['x_prime']

    proposal.configure_reparameterisations = MagicMock()
    proposal.configure_reparameterisations.side_effect = update

    FlowProposal.set_rescaling(proposal)

    proposal.set_boundary_inversion.assert_called_once()
    proposal.configure_reparameterisations.assert_called_once_with(
        {'x': 'default'}
    )
    assert proposal.reparameterisations == {'x': 'default'}
    assert proposal.rescaled_names == ['x_prime']


def test_set_rescaling_with_reparameterisations(proposal, model):
    """
    Test setting the rescaling when a reparameterisations dict is defined.
    """
    proposal.model = model
    proposal.model.reparameterisations = None
    proposal.reparameterisations = {'x': 'default'}
    proposal.set_boundary_inversion = MagicMock()

    def update(self):
        proposal.rescale_parameters = ['x']
        proposal.rescaled_names = ['x_prime']

    proposal.configure_reparameterisations = MagicMock()
    proposal.configure_reparameterisations.side_effect = update

    FlowProposal.set_rescaling(proposal)

    proposal.set_boundary_inversion.assert_called_once()
    proposal.configure_reparameterisations.assert_called_once_with(
        {'x': 'default'}
    )
    assert proposal.reparameterisations == {'x': 'default'}
    assert proposal.rescaled_names == ['x_prime']


@pytest.mark.parametrize('update_bounds', [True, False])
@pytest.mark.parametrize(
    'rescale_parameters',
    [
        (False, ['x', 'y']),
        (True, ['x_prime', 'y_prime']),
        (['x'], ['x_prime', 'y'])
    ]
)
def test_set_rescaling_parameters_list(proposal, model, update_bounds,
                                       rescale_parameters):
    """Test setting rescaling without reparameterisations."""
    proposal.model = model
    proposal.model.reparameterisations = None
    proposal.reparameterisations = None
    proposal.configure_reparameterisations = MagicMock()

    proposal.rescale_parameters = rescale_parameters[0]
    proposal.rescale_bounds = [-1, 1]
    proposal.update_bounds = update_bounds

    FlowProposal.set_rescaling(proposal)

    proposal.configure_reparameterisations.assert_not_called()
    assert proposal.rescaled_names == rescale_parameters[1]

    if rescale_parameters[0]:
        assert proposal._rescale_factor == 2.0
        assert proposal._rescale_shift == -1.0
        assert proposal._min == {'x': -5, 'y': -5}
        assert proposal._max == {'x': 5, 'y': 5}


def test_default_rescale(proposal):
    """Test the default rescaling method."""
    x = np.array([[1, 2], [3, 4]])
    x_prime, log_j = FlowProposal.rescale(proposal, x)
    np.testing.assert_array_equal(x, x_prime)
    assert (log_j == 0.).all()


def test_default_inverse_rescale(proposal):
    """Test the default inverse rescaling method."""
    x_prime = np.array([[1, 2], [3, 4]])
    x, log_j = FlowProposal.inverse_rescale(proposal, x_prime)
    np.testing.assert_array_equal(x, x_prime)
    assert (log_j == 0.).all()


@pytest.mark.parametrize('n', [1, 10])
def test_rescale_w_reparameterisation(proposal, n):
    """Test rescaling when using reparameterisation dict"""
    x = numpy_array_to_live_points(np.random.randn(n, 2), ['x', 'y'])
    x['logL'] = np.random.randn(n)
    x['logP'] = np.random.randn(n)
    x_prime = numpy_array_to_live_points(
        np.random.randn(n, 2), ['x_prime', 'y_prime'])
    proposal.x_prime_dtype = \
        [('x_prime', 'f8'), ('y_prime', 'f8'), ('logP', 'f8'), ('logL', 'f8')]
    proposal._reparameterisation = MagicMock()
    proposal._reparameterisation.reparameterise = MagicMock(return_value=[
        x, x_prime, np.ones(x.size)])

    x_prime_out, log_j = \
        FlowProposal._rescale_w_reparameterisation(
            proposal, x, compute_radius=False, test='lower')

    np.testing.assert_array_equal(
        x_prime[['x_prime', 'y_prime']], x_prime_out[['x_prime', 'y_prime']])
    np.testing.assert_array_equal(
        x[['logP', 'logL']], x_prime_out[['logL', 'logP']])
    proposal._reparameterisation.reparameterise.assert_called_once()


@pytest.mark.parametrize('n', [1, 10])
def test_inverse_rescale_w_reparameterisation(proposal, n):
    """Test rescaling when using reparameterisation dict"""
    x = numpy_array_to_live_points(np.random.randn(n, 2), ['x', 'y']).squeeze()
    x_prime = numpy_array_to_live_points(
        np.random.randn(n, 2), ['x_prime', 'y_prime'])
    x_prime['logL'] = np.random.randn(n)
    x_prime['logP'] = np.random.randn(n)
    proposal.x_dtype = \
        [('x', 'f8'), ('y', 'f8'), ('logP', 'f8'), ('logL', 'f8')]
    proposal._reparameterisation = MagicMock()
    proposal._reparameterisation.inverse_reparameterise = \
        MagicMock(return_value=[x, x_prime, np.ones(x.size)])

    x_out, log_j = \
        FlowProposal._inverse_rescale_w_reparameterisation(
            proposal, x_prime)

    np.testing.assert_array_equal(x[['x', 'y']], x_out[['x', 'y']])
    np.testing.assert_array_equal(
        x_prime[['logP', 'logL']], x_out[['logL', 'logP']])
    proposal._reparameterisation.inverse_reparameterise.assert_called_once()


@pytest.mark.parametrize('n', [1, 10])
@pytest.mark.parametrize('compute_radius', [False, True])
def test_rescale_to_bounds(proposal, model, n, compute_radius):
    """Test the default rescaling to bounds.

    Also tests the log Jacobian determinant.
    """
    model.names.append('z')
    model.bounds['z'] = [-10, 10]
    x = numpy_array_to_live_points(
        np.random.randn(n, 3), ['x', 'y', 'z']
    ).squeeze()
    x_prime_expected = numpy_array_to_live_points(
        np.zeros([n, 3]), ['x_prime', 'y_prime', 'z'])

    x_prime_expected['x_prime'] = 2 * (x['x'] + 5) / 10 - 1
    x_prime_expected['y_prime'] = 2 * (x['y'] + 4) / 8 - 1
    x_prime_expected['z'] = x['z']

    proposal.x_prime_dtype = \
        [('x_prime', 'f8'), ('y_prime', 'f8'), ('z', 'f8'), ('logP', 'f8'),
         ('logL', 'f8')]

    proposal.names = ['x', 'y', 'z']
    proposal.rescale_parameters = ['x', 'y']
    proposal.rescaled_names = ['x_prime', 'y_prime', 'z']
    proposal.model = model
    proposal.boundary_inversion = []

    proposal._rescale_factor = 2.0
    proposal._rescale_shift = -1.0
    proposal._min = {'x': -5, 'y': -4}
    proposal._max = {'x': 5, 'y': 4}

    x_prime, log_j = \
        FlowProposal._rescale_to_bounds(
            proposal, x, compute_radius=compute_radius)

    np.testing.assert_equal(log_j, -np.log(20))
    np.testing.assert_array_equal(x_prime, x_prime_expected)


@pytest.mark.parametrize('n', [1, 10])
@pytest.mark.parametrize('inputs', [('split', True), ('duplicate', False)])
@pytest.mark.parametrize('bound', ['lower', 'upper'])
def test_rescale_to_bounds_w_inversion_duplicate(
    proposal,
    model,
    n,
    inputs,
    bound
):
    """Test the default rescaling with inversion using duplication."""
    x = numpy_array_to_live_points(np.random.randn(n, 2), ['x', 'y']).squeeze()
    x_prime_expected = numpy_array_to_live_points(
        np.zeros([n, 2]), ['x_prime', 'y_prime'])

    x_prime_expected['x_prime'] = (x['x'] + 5) / 10
    x_prime_expected['y_prime'] = (x['y'] + 4) / 8

    if bound == 'lower':
        x_prime_expected = np.concatenate([x_prime_expected, x_prime_expected])
        x_prime_expected['x_prime'][n:] *= -1
    elif bound == 'upper':
        x_prime_expected['x_prime'] = 1 - x_prime_expected['x_prime']
        x_prime_expected = np.concatenate([x_prime_expected, x_prime_expected])
        x_prime_expected['x_prime'][n:] *= -1

    proposal.x_prime_dtype = \
        [('x_prime', 'f8'), ('y_prime', 'f8'), ('logP', 'f8'), ('logL', 'f8')]

    proposal.names = ['x', 'y']
    proposal.rescale_parameters = ['x', 'y']
    proposal.rescaled_names = ['x_prime', 'y_prime']
    proposal.model = model
    proposal.boundary_inversion = ['x']

    proposal._rescale_factor = 1.0
    proposal._rescale_shift = 0.0
    proposal._min = {'x': -5, 'y': -4}
    proposal._max = {'x': 5, 'y': 4}
    proposal._edges = {'x': None}
    proposal.detect_edges_kwargs = {'k': 2}
    proposal.inversion_type = inputs[0]

    with patch('nessai.proposal.flowproposal.detect_edge',
               return_value=bound) as mock_detect_edge:
        x_prime, log_j = \
            FlowProposal._rescale_to_bounds(
                proposal, x, compute_radius=inputs[1], test=True)

    np.testing.assert_array_equal(
        mock_detect_edge.call_args[0][0],
        x_prime_expected['x_prime'][:n]
    )
    assert mock_detect_edge.call_args[1]['test'] is True
    assert mock_detect_edge.call_args[1]['k'] == 2

    np.testing.assert_equal(log_j, -np.log(80))
    np.testing.assert_array_equal(x_prime, x_prime_expected)


@pytest.mark.parametrize('n', [1, 10])
@pytest.mark.parametrize('bound', ['lower', 'upper'])
def test_rescale_to_bounds_w_inversion_split(
    proposal,
    model,
    n,
    bound
):
    """Test the default rescaling with inversion using splitting."""
    x = numpy_array_to_live_points(np.random.randn(n, 2), ['x', 'y']).squeeze()
    x_prime_expected = numpy_array_to_live_points(
        np.zeros([n, 2]), ['x_prime', 'y_prime'])

    x_prime_expected['x_prime'] = (x['x'] + 5) / 10
    x_prime_expected['y_prime'] = (x['y'] + 4) / 8

    inv = np.random.choice(n, n // 2, replace=False)

    if bound == 'lower':
        x_prime_expected['x_prime'][inv] *= -1
    elif bound == 'upper':
        x_prime_expected['x_prime'] = 1 - x_prime_expected['x_prime']
        x_prime_expected['x_prime'][inv] *= -1

    proposal.x_prime_dtype = \
        [('x_prime', 'f8'), ('y_prime', 'f8'), ('logP', 'f8'), ('logL', 'f8')]

    proposal.names = ['x', 'y']
    proposal.rescale_parameters = ['x', 'y']
    proposal.rescaled_names = ['x_prime', 'y_prime']
    proposal.model = model
    proposal.boundary_inversion = ['x']

    proposal._rescale_factor = 1.0
    proposal._rescale_shift = 0.0
    proposal._min = {'x': -5, 'y': -4}
    proposal._max = {'x': 5, 'y': 4}
    proposal._edges = {'x': None}
    proposal.detect_edges_kwargs = {'k': 2}
    proposal.inversion_type = 'split'

    with patch('nessai.proposal.flowproposal.detect_edge',
               return_value=bound) as mock_detect_edge, \
         patch('numpy.random.choice', return_value=inv):
        x_prime, log_j = \
            FlowProposal._rescale_to_bounds(
                proposal, x, compute_radius=False, test=True)

    np.testing.assert_array_equal(
        mock_detect_edge.call_args[0][0],
        x_prime_expected['x_prime'][:n]
    )
    assert mock_detect_edge.call_args[1]['test'] is True
    assert mock_detect_edge.call_args[1]['k'] == 2

    np.testing.assert_equal(log_j, -np.log(80))
    np.testing.assert_array_equal(x_prime, x_prime_expected)


@pytest.mark.parametrize('n', [1, 10])
@pytest.mark.parametrize('bound', ['lower', 'upper'])
def test_rescale_to_bounds_w_inversion_false(
    proposal,
    model,
    n,
    bound
):
    """
    Test the default rescaling without inversion and with edges already set.
    """
    x = numpy_array_to_live_points(np.random.randn(n, 2), ['x', 'y']).squeeze()
    x_prime_expected = numpy_array_to_live_points(
        np.zeros([n, 2]), ['x_prime', 'y_prime'])

    x_prime_expected['x_prime'] = (x['x'] + 5) / 10
    x_prime_expected['y_prime'] = (x['y'] + 4) / 8

    proposal.x_prime_dtype = \
        [('x_prime', 'f8'), ('y_prime', 'f8'), ('logP', 'f8'), ('logL', 'f8')]

    proposal.names = ['x', 'y']
    proposal.rescale_parameters = ['x', 'y']
    proposal.rescaled_names = ['x_prime', 'y_prime']
    proposal.model = model
    proposal.boundary_inversion = ['x']

    proposal._rescale_factor = 1.0
    proposal._rescale_shift = 0.0
    proposal._min = {'x': -5, 'y': -4}
    proposal._max = {'x': 5, 'y': 4}
    proposal._edges = {'x': False}
    proposal.detect_edges_kwargs = {'k': 2}
    proposal.inversion_type = 'split'

    with patch('nessai.proposal.flowproposal.detect_edge',
               return_value=bound) as mock_detect_edge:
        x_prime, log_j = \
            FlowProposal._rescale_to_bounds(
                proposal, x, compute_radius=False, test=True)

    mock_detect_edge.assert_not_called()

    np.testing.assert_equal(log_j, -np.log(80))
    np.testing.assert_array_equal(x_prime, x_prime_expected)


@pytest.mark.parametrize('n', [1, 10])
def test_inverse_rescale_to_bounds(proposal, model, n):
    """Test the default method for the inverse rescaling.

    Also tests the log Jacobian determinant.
    """
    model.names.append('z')
    model.bounds['z'] = [-10, 10]
    x_prime = numpy_array_to_live_points(
        np.random.randn(n, 3),
        ['x_prime', 'y_prime', 'z']
    ).squeeze()
    x_expected = \
        numpy_array_to_live_points(np.zeros([n, 3]), ['x', 'y', 'z'])

    x_expected['x'] = 10.0 * (x_prime['x_prime'] + 1.0) / 2.0 - 5.0
    x_expected['y'] = 8.0 * (x_prime['y_prime'] + 1.0) / 2.0 - 4.0
    x_expected['z'] = x_prime['z']

    proposal.x_dtype = \
        [('x', 'f8'), ('y', 'f8'), ('z', 'f8'), ('logP', 'f8'), ('logL', 'f8')]

    proposal.names = ['x', 'y', 'z']
    proposal.rescale_parameters = ['x', 'y']
    proposal.rescaled_names = ['x_prime', 'y_prime', 'z']
    proposal.model = model
    proposal.boundary_inversion = []

    proposal._rescale_factor = 2.0
    proposal._rescale_shift = -1.0
    proposal._min = {'x': -5, 'y': -4}
    proposal._max = {'x': 5, 'y': 4}

    x, log_j = \
        FlowProposal._inverse_rescale_to_bounds(proposal, x_prime)

    np.testing.assert_equal(log_j, np.log(20))
    np.testing.assert_array_equal(x, x_expected)


@pytest.mark.parametrize('n', [1, 10])
@pytest.mark.parametrize('itype', ['lower', 'upper'])
def test_inverse_rescale_to_bounds_w_inversion(proposal, model, n, itype):
    """Test the default method for the inverse rescaling with inversion.

    Also tests the log Jacobian determinant.
    """
    model.names.append('z')
    model.bounds['z'] = [-10, 10]
    x_prime = numpy_array_to_live_points(
        np.random.uniform(-1, 1, (n, 3)),
        ['x_prime', 'y_prime', 'z']
    ).squeeze()

    x_expected = \
        numpy_array_to_live_points(np.zeros([n, 3]), ['x', 'y', 'z'])

    if itype == 'lower':
        x_expected['x'] = 10.0 * np.abs(x_prime['x_prime']) - 5.0
    elif itype == 'upper':
        x_expected['x'] = 10.0 * (1 - np.abs(x_prime['x_prime'])) - 5.0

    x_expected['y'] = 8.0 * x_prime['y_prime'] - 4.0
    x_expected['z'] = x_prime['z']

    proposal.x_dtype = \
        [('x', 'f8'), ('y', 'f8'), ('z', 'f8'), ('logP', 'f8'), ('logL', 'f8')]

    proposal.names = ['x', 'y', 'z']
    proposal.rescale_parameters = ['x', 'y']
    proposal.rescaled_names = ['x_prime', 'y_prime', 'z']
    proposal.model = model
    proposal.boundary_inversion = ['x']
    proposal.inversion_type = 'split'
    proposal._edges = {'x': itype}

    proposal._rescale_factor = 1.0
    proposal._rescale_shift = 0.0
    proposal._min = {'x': -5, 'y': -4}
    proposal._max = {'x': 5, 'y': 4}

    x, log_j = \
        FlowProposal._inverse_rescale_to_bounds(proposal, x_prime)

    np.testing.assert_equal(log_j, np.log(80))
    np.testing.assert_array_equal(x, x_expected)


@pytest.mark.parametrize('has_inversion', [False, True])
def test_verify_rescaling(proposal, has_inversion):
    """Test the method that tests the rescaling at runtime"""
    x = np.array([[1], [2]], dtype=[('x', 'f8')])
    x_prime = x['x'] / 2
    log_j = np.array([-2, -2])
    x_out = x.copy()
    log_j_inv = np.array([2, 2])

    if has_inversion:
        x_prime = np.concatenate([x_prime, x_prime])
        log_j = np.concatenate([log_j, log_j])
        log_j_inv = np.concatenate([log_j_inv, log_j_inv])
        x_out = np.concatenate([x_out, x_out])

    proposal.model = MagicMock()
    proposal.model.new_point = MagicMock(return_value=x)
    proposal.rescale = MagicMock(return_value=(x_prime, log_j))
    proposal.inverse_rescale = MagicMock(return_value=(x_out, log_j_inv))
    proposal.check_state = MagicMock()

    FlowProposal.verify_rescaling(proposal)

    proposal.check_state.assert_has_calls(4 * [call(x)])
    # Should call 4 different test cases
    calls = [
        call(x, test='lower'),
        call(x, test='upper'),
        call(x, test=False),
        call(x, test=None)
    ]
    proposal.rescale.assert_has_calls(calls)
    proposal.inverse_rescale.assert_has_calls(4 * [call(x_prime)])


@pytest.mark.parametrize('has_inversion', [False, True])
def test_verify_rescaling_invertible_error(proposal, has_inversion):
    """Assert an error is raised if the rescaling is not invertible"""
    x = np.array([[1], [2]], dtype=[('x', 'f8')])
    x_prime = x['x'] / 2
    log_j = np.array([-2, -2])
    x_out = x.copy()[::-1]
    log_j_inv = np.array([2, 2])

    if has_inversion:
        x_prime = np.concatenate([x_prime, x_prime])
        log_j = np.concatenate([log_j, log_j])
        log_j_inv = np.concatenate([log_j_inv, log_j_inv])
        x_out = np.concatenate([x_out, x_out])

    proposal.model = MagicMock()
    proposal.model.new_point = MagicMock(return_value=x)
    proposal.rescale = MagicMock(return_value=(x_prime, log_j))
    proposal.inverse_rescale = MagicMock(return_value=(x_out, log_j_inv))

    with pytest.raises(RuntimeError) as excinfo:
        FlowProposal.verify_rescaling(proposal)
    assert 'Rescaling is not invertible for x' in str(excinfo.value)


def test_verify_rescaling_duplicate_error(proposal):
    """Assert an error is raised if the duplication is missing samples"""
    x = np.array([[1], [2]], dtype=[('x', 'f8')])
    x_prime = x['x'] / 2
    log_j = np.array([-2, -2, -2, -2])
    x_out = np.array([[1], [3], [4], [5]], dtype=[('x', 'f8')])
    log_j_inv = np.array([2, 2, 2, 2])

    proposal.model = MagicMock()
    proposal.model.new_point = MagicMock(return_value=x)
    proposal.rescale = MagicMock(return_value=(x_prime, log_j))
    proposal.inverse_rescale = MagicMock(return_value=(x_out, log_j_inv))

    with pytest.raises(RuntimeError) as excinfo:
        FlowProposal.verify_rescaling(proposal)
    assert 'Duplicate samples must map to same input' in str(excinfo.value)


@pytest.mark.parametrize('has_inversion', [False, True])
def test_verify_rescaling_jacobian_error(proposal, has_inversion):
    """Assert an error is raised if the Jacobian is not invertible"""
    x = np.array([[1], [2]], dtype=[('x', 'f8')])
    x_prime = x['x'] / 2
    log_j = np.array([-2, -2])
    x_out = x.copy()
    log_j_inv = np.array([2, 1])

    if has_inversion:
        x_prime = np.concatenate([x_prime, x_prime])
        log_j = np.concatenate([log_j, log_j])
        log_j_inv = np.concatenate([log_j_inv, log_j_inv])
        x_out = np.concatenate([x_out, x_out])

    proposal.model = MagicMock()
    proposal.model.new_point = MagicMock(return_value=x)
    proposal.rescale = MagicMock(return_value=(x_prime, log_j))
    proposal.inverse_rescale = MagicMock(return_value=(x_out, log_j_inv))

    with pytest.raises(RuntimeError) as excinfo:
        FlowProposal.verify_rescaling(proposal)
    assert 'Rescaling Jacobian is not invertible' in str(excinfo.value)


def test_check_state_boundary_inversion_default(proposal):
    """Test the check state method for boundary inversion"""
    x = numpy_array_to_live_points(np.random.randn(10, 2), ['x', 'y'])
    proposal._reparameterisation = None
    proposal._edges = {'x': 'lower', 'y': 'upper'}
    proposal.boundary_inversion = ['x', 'y']
    proposal.update_bounds = False
    FlowProposal.check_state(proposal, x)
    assert all(v is None for v in proposal._edges.values())


def test_check_state_boundary_inversion_reparameterisations(proposal):
    """
    Test the check state method for boundary inversion with
    reparameterisations.
    """
    x = numpy_array_to_live_points(np.random.randn(10, 2), ['x', 'y'])
    proposal._reparameterisation = Mock()
    proposal._reparameterisation.reset_inversion = MagicMock()
    proposal.boundary_inversion = ['x', 'y']
    proposal.update_bounds = False
    FlowProposal.check_state(proposal, x)
    proposal._reparameterisation.reset_inversion.assert_called_once()


def test_check_state_update_bounds_default(proposal, model):
    """
    Test the check state method for updating bounds.
    """
    x = numpy_array_to_live_points(np.random.randn(10, 2), ['x', 'y'])
    proposal.model = model
    excepted_min = (np.min(x['x']), np.min(x['y']))
    excepted_max = (np.max(x['x']), np.max(x['y']))
    proposal._reparameterisation = None
    proposal.boundary_inversion = False
    proposal.update_bounds = True
    proposal._min = {'x': -np.inf, 'y': -np.inf}
    proposal._max = {'x': np.inf, 'y': np.inf}
    FlowProposal.check_state(proposal, x)

    assert proposal._min['x'] == excepted_min[0]
    assert proposal._min['y'] == excepted_min[1]
    assert proposal._max['x'] == excepted_max[0]
    assert proposal._max['y'] == excepted_max[1]


def test_check_state_update_bounds_reparameterisations(proposal):
    """
    Test the check state method for updating bounds.
    """
    x = numpy_array_to_live_points(np.random.randn(10, 2), ['x', 'y'])
    proposal._reparameterisation = Mock()
    proposal._reparameterisation.update_bounds = MagicMock()
    proposal.boundary_inversion = False
    proposal.update_bounds = True

    FlowProposal.check_state(proposal, x)
    proposal._reparameterisation.update_bounds.assert_called_once()



# -*- coding: utf-8 -*-
"""
Test the multiprocessing pool for the different proposal methods.
"""
from numpy.testing import assert_array_equal
import pytest

from nessai.proposal import (
    AnalyticProposal,
    AugmentedFlowProposal,
    FlowProposal,
    RejectionProposal
)


@pytest.mark.parametrize(
    'proposal',
    [AnalyticProposal, AugmentedFlowProposal, FlowProposal, RejectionProposal]
)
@pytest.mark.parametrize('n_pool', [None, 1])
@pytest.mark.timeout(10)
@pytest.mark.flaky(run=3)
@pytest.mark.integration_test
def test_analytic_pool(proposal, n_pool, model, tmpdir):
    try:
        prop = proposal(model, n_pool=n_pool, poolsize=2, plot=False,
                        max_radius=1, output=str(tmpdir.mkdir('test')))
    except TypeError:
        prop = proposal(model, n_pool=n_pool, poolsize=2)

    prop.initialise()

    prop.configure_pool()
    if n_pool:
        assert prop.pool
    else:
        assert prop.pool is None

    worst = model.new_point()
    prop.draw(worst)

    if n_pool:
        logL = [model.log_likelihood(s) for s in prop.samples]
        assert_array_equal(prop.samples['logL'], logL)
    else:
        assert all(ll == 0. for ll in prop.samples['logL'])

    prop.close_pool()
    assert prop.pool is None



#!/usr/bin/env python

# Example of parallelising the likelihood evaluation in nessai

import numpy as np

from nessai.flowsampler import FlowSampler
from nessai.model import Model
from nessai.utils import setup_logger


output = './outdir/parallelisation_example/'
logger = setup_logger(output=output, log_level='WARNING')


class GaussianLikelihood(Model):
    """
    Gaussian likelihood with the mean and standard deviation as the parameters
    to infer.

    Parameters
    ----------
    n_points : int, optional
        Number of points to sample for the data. More points will lead to
        a slower likelihood.
    """
    def __init__(self, n_points=1000):
        self.names = ['mu', 'sigma']
        self.bounds = {'mu': [-3, 3], 'sigma': [0.01, 3]}
        self.truth = {'mu': 1.7, 'sigma': 0.7}

        self.data = np.random.normal(
            self.truth['mu'], self.truth['sigma'], size=n_points
        )

    def log_prior(self, x):
        """
        Uniform prior on both parameters.
        """
        log_p = 0.
        for n in self.names:
            log_p += (np.log((x[n] >= self.bounds[n][0])
                             & (x[n] <= self.bounds[n][1]))
                      - np.log(self.bounds[n][1] - self.bounds[n][0]))
        return log_p

    def log_likelihood(self, x):
        """
        Gaussian likelihood.
        """
        log_l = np.sum(
            - np.log(x['sigma']) -
            0.5 * ((self.data - x['mu']) / x['sigma']) ** 2
        )
        return log_l


flow_config = dict(
    batch_size=1000,
    max_epochs=200,
    patience=20,
    model_config=dict(n_blocks=2, n_neurons=4, n_layers=2)
)

# Configure the sampler with 3 total threads, 2 of which are used for
# evaluating the likelihood.
fs = FlowSampler(
    GaussianLikelihood(),
    output=output,
    flow_config=flow_config,
    resume=False,
    seed=1234,
    max_threads=3,               # Maximum number of threads
    n_pool=2,                    # Threads for evaluating the likelihood
    nlive=2000,
    maximum_uninformed=2000,
    proposal_plots=False,
)

# Run the sampler
fs.run()



# -*- coding: utf-8 -*-
"""
Tests related to checking and updating the state of sampler and the history.
"""
import pytest
from unittest.mock import MagicMock, patch

from nessai.nestedsampler import NestedSampler


@pytest.fixture
def sampler(sampler):
    """A sampler mock configured to work with update state"""
    # Stored stats
    sampler.population_acceptance = [0.5]
    sampler.population_radii = [1.0]
    sampler.population_iterations = [0]
    sampler.population_acceptance = [0.5]

    sampler.likelihood_evaluations = [10]
    sampler.min_likelihood = [0.0]
    sampler.max_likelihood = [100.0]
    sampler.logZ_history = [-100.0]
    sampler.dZ_history = [100.0]
    sampler.mean_acceptance_history = [0.5]
    # Attributed used to update stats
    sampler.model = MagicMock()
    sampler.model.likelihood_evaluations = 100
    sampler.logLmin = 0.0
    sampler.logLmax = 150.0
    sampler.condition = 50.0

    sampler.state = MagicMock()
    sampler.state.logZ = -50.0
    sampler.state.info = [0.0]

    sampler.proposal = MagicMock()
    sampler.proposal.r = 2.0
    sampler.proposal.population_acceptance = 0.4

    sampler.mean_acceptance = 0.5
    sampler.mean_block_acceptance = 0.5
    sampler.block_acceptance = 0.5
    sampler.block_iteration = 5

    sampler.checkpoint = MagicMock()

    return sampler


@pytest.mark.parametrize('switch', [False, True])
@pytest.mark.parametrize('uninformed', [False, True])
def test_check_state_force(sampler, switch, uninformed):
    """Test the behaviour of check_state with force=True.

    Training should always start irrespective of other checks and with
    force=True unless uninformed sampling is being used and the switch=False.
    """
    sampler.uninformed_sampling = uninformed
    sampler.check_proposal_switch = MagicMock(return_value=switch)
    sampler.check_training = MagicMock()
    sampler.train_proposal = MagicMock()

    NestedSampler.check_state(sampler, force=True)

    if uninformed and not switch:
        sampler.train_proposal.assert_not_called()
    else:
        sampler.train_proposal.assert_called_once_with(force=True)

    sampler.check_training.assert_not_called()


@pytest.mark.parametrize('force', [False, True])
@pytest.mark.parametrize('train', [False, True])
def test_check_state_train(sampler, force, train):
    """Test the behaviour of check_state with force=False and `check_training`
    returns True, False, True, True, False, True, or False, False.

    Force is used for the return values for `check_training`
    """
    sampler.uninformed_sampling = False
    sampler.check_proposal_switch = MagicMock()
    sampler.check_training = MagicMock(return_value=(train, force))
    sampler.train_proposal = MagicMock()

    NestedSampler.check_state(sampler, force=False)
    if train or force:
        sampler.check_training.assert_called_once_with()
        sampler.train_proposal.assert_called_once_with(force=force)
    else:
        sampler.check_training.assert_called_once_with()
        sampler.train_proposal.assert_not_called()


@patch('nessai.nestedsampler.NestedSampler.checkpoint')
def test_update_state_checked_acceptance(mock, sampler):
    """Test the behaviour of update state if `_checked_population` is False.

    Checks to make sure the correct statistics are stored and the other
    checks are not called because it / nlive and it / (nlive/10) != 0
    """
    sampler.iteration = 11
    sampler.proposal._checked_population = False

    NestedSampler.update_state(sampler)

    assert sampler.population_acceptance == [0.5, 0.4]
    assert sampler.population_radii == [1.0, 2.0]
    assert sampler.population_iterations == [0, 11]
    assert sampler.proposal._checked_population is True
    assert sampler.likelihood_evaluations == [10]
    assert not mock.called


def test_update_state_history(sampler):
    """Test the behaviour of updated state if it / (nlive/10) == 0

    Checks to make sure the correct statistics are saved and metrics related to
    population are not updated.
    """
    sampler.iteration = 10
    sampler.proposal._checked_population = True

    NestedSampler.update_state(sampler)

    assert sampler.population_acceptance == [0.5]
    assert sampler.likelihood_evaluations == [10, 100]
    assert sampler.min_likelihood == [0.0, 0.0]
    assert sampler.max_likelihood == [100.0, 150.0]
    assert sampler.logZ_history == [-100.0, -50.0]
    assert sampler.dZ_history == [100.0, 50.0]
    assert sampler.mean_acceptance_history == [0.5, 0.5]
    assert not sampler.checkpoint.called

    assert sampler.proposal.ns_acceptance == 0.5


@pytest.mark.parametrize('checkpointing', [False, True])
@pytest.mark.parametrize('plot', [False, True])
@patch('nessai.nestedsampler.plot_indices')
def test_update_state_every_nlive(mock_plot, plot, checkpointing, sampler):
    """Test the update that happens every nlive iterations.

    Tests both with plot=True and plot=False
    """
    sampler.nlive = 100
    sampler.iteration = 100
    sampler.proposal._checked_population = True
    sampler.check_insertion_indices = MagicMock()
    sampler.plot = plot
    sampler.uninformed_sampling = True
    sampler.plot_state = MagicMock()
    sampler.plot_trace = MagicMock()
    sampler.output = './'
    sampler.insertion_indices = range(2 * sampler.nlive)
    sampler.checkpointing = checkpointing

    NestedSampler.update_state(sampler)

    if checkpointing:
        sampler.checkpoint.assert_called_once_with(periodic=True)
    else:
        sampler.checkpoint.assert_not_called()
    sampler.check_insertion_indices.assert_called_once()
    assert sampler.block_iteration == 0
    assert sampler.block_acceptance == 0.
    assert sampler.likelihood_evaluations == [10, 100]
    assert sampler.population_acceptance == [0.5]

    if plot:
        sampler.plot_state.assert_called_once_with(filename='.//state.png')
        sampler.plot_trace.assert_called_once()
        mock_plot.assert_called_once_with(
            sampler.insertion_indices[-100:], 100, plot_breakdown=False,
            filename='.//diagnostics/insertion_indices_100.png')
    else:
        assert not sampler.plot_state.called
        assert not sampler.plot_trace.called
        assert not mock_plot.called


@pytest.mark.parametrize('checkpointing', [False, True])
@patch('nessai.nestedsampler.plot_indices')
def test_update_state_force(mock_plot, checkpointing, sampler):
    """Test the update that happens if force=True.

    Checks that plot_indices is not called even if plotting is enabled.
    """
    sampler.iteration = 111
    sampler.proposal._checked_population = True
    sampler.check_insertion_indices = MagicMock()
    sampler.plot = True
    sampler.plot_state = MagicMock()
    sampler.plot_trace = MagicMock()
    sampler.output = './'
    sampler.uninformed_sampling = False
    sampler.checkpointing = checkpointing

    NestedSampler.update_state(sampler, force=True)

    if checkpointing:
        sampler.checkpoint.assert_called_once_with(periodic=True)
    else:
        sampler.checkpoint.assert_not_called()
    assert not mock_plot.called
    assert not sampler.called
    sampler.plot_trace.assert_called_once()
    sampler.plot_state.assert_called_once_with(filename='.//state.png')

    assert sampler.max_likelihood == [100.0, 150.0]
    assert sampler.population_acceptance == [0.5]
    assert sampler.block_acceptance == 0.5
    assert sampler.block_iteration == 5



# -*- coding: utf-8 -*-
"""
Test the base flow class
"""
import pytest
from unittest.mock import create_autospec, patch

from nessai.flows import BaseFlow, NFlow
from nflows.transforms import Transform


@pytest.fixture
def flow():
    return create_autospec(BaseFlow)


def test_base_flow_abstract_methods():
    """Assert an error is raised if the methods are not implemented."""
    with pytest.raises(TypeError) as excinfo:
        BaseFlow()
    assert 'instantiate abstract class BaseFlow with abstract method' \
        in str(excinfo.value)


@pytest.mark.parametrize(
    'method',
    ['forward', 'inverse', 'sample', 'log_prob', 'base_distribution_log_prob',
     'forward_and_log_prob', 'sample_and_log_prob']
)
def test_base_flow_methods(method, flow):
    """Test to make sure all of the method in the base class raise errors"""
    with pytest.raises(NotImplementedError):
        getattr(BaseFlow, method)(flow, None)


def test_base_flow_to(flow):
    """Assert the to method works as intended"""
    flow.device = None
    with patch('nessai.flows.base.Module.to') as super_init:
        BaseFlow.to(flow, 'cpu')
    super_init.assert_called_once_with('cpu')
    assert flow.device == 'cpu'


def test_nflow_base_transform():
    """Assert an error is raised if the transform class is invalid."""
    class Test:
        pass

    with pytest.raises(TypeError) as excinfo:
        NFlow(Test(), Test())
    assert 'transform must inherit' in str(excinfo.value)


def test_nflow_base_distribution():
    """Assert an error is raised if the distribution class is invalid."""
    class Test:
        pass

    with pytest.raises(TypeError) as excinfo:
        NFlow(Transform(), Test())
    assert 'distribution must inherit' in str(excinfo.value)




# -*- coding: utf-8 -*-
"""
Test the FlowModel object.
"""
import numpy as np
import pickle
import pytest
import torch
from unittest.mock import create_autospec, MagicMock, patch

from nessai.flowmodel import update_config, FlowModel


@pytest.fixture()
def data_dim():
    return 2


@pytest.fixture()
def model():
    return create_autospec(FlowModel)


@pytest.fixture(scope='function')
def flow_model(flow_config, data_dim, tmpdir):
    flow_config['model_config']['n_inputs'] = data_dim
    output = str(tmpdir.mkdir('flowmodel'))
    return FlowModel(flow_config, output=output)


def test_update_config_none():
    """Test update config for no input"""
    config = update_config(None)
    assert 'model_config' in config


def test_update_config_invalid_type():
    """Test update config when an invalid argument is specified"""
    with pytest.raises(TypeError) as excinfo:
        update_config(False)
    assert 'Must pass a dictionary' in str(excinfo.value)


@pytest.mark.parametrize('noise_scale', ['auto', 4])
def test_update_config_invalid_noise_scale(noise_scale):
    """Assert an error is raised if noise_scale is not a float or adapative."""
    config = {'noise_scale': noise_scale}
    with pytest.raises(ValueError) as excinfo:
        update_config(config)
    assert 'noise_scale must be a float or' in str(excinfo.value)


def test_init_no_config(tmpdir):
    """Test the init method with no config specified"""
    output = str(tmpdir.mkdir('no_config'))
    default_config = update_config(None)
    fm = FlowModel(output=output)

    assert fm.model_config == default_config['model_config']


def test_init_config_class(tmpdir):
    """Test the init and save methods when specifying `flow` as a class"""
    from nessai.flows import RealNVP
    output = str(tmpdir.mkdir('no_config'))
    config = dict(model_config=dict(flow=RealNVP))
    fm = FlowModel(config=config, output=output)

    assert fm.model_config['flow'].__name__ == 'RealNVP'


def test_initialise(model):
    """Test the initialise method"""
    model.get_optimiser = MagicMock()
    model.model_config = dict(n_neurons=2)
    model.inference_device = None
    model.optimiser = 'adam'
    model.optimiser_kwargs = {'weights': 0.1}
    with patch('nessai.flowmodel.configure_model',
               return_value=('model', 'cpu')) as mock:
        FlowModel.initialise(model)
    mock.assert_called_once_with(model.model_config)
    model.get_optimiser.assert_called_once_with('adam', weights=0.1)
    assert model.inference_device == torch.device('cpu')


@pytest.mark.parametrize('optimiser', ['Adam', 'AdamW', 'SGD'])
def test_get_optimiser(model, optimiser):
    """Test to make sure the three supported optimisers work"""
    model.lr = 0.1
    model.model = MagicMock()
    with patch(f'torch.optim.{optimiser}') as mock:
        FlowModel.get_optimiser(model, optimiser)
    mock.assert_called_once()


@pytest.mark.parametrize('val_size, batch_size', [(0.1, 100),
                                                  (0.5, 'all')])
def test_prep_data(flow_model, data_dim, val_size, batch_size):
    """
    Test the data prep, make sure batch sizes and validation size
    produce the correct result.
    """
    n = 100
    x = np.random.randn(n, data_dim)

    train, val = flow_model.prep_data(x, val_size, batch_size)
    if batch_size == 'all':
        batch_size = int(n * (1 - val_size))

    assert flow_model.batch_size == batch_size
    assert len(train) + len(val) == n


@pytest.mark.parametrize('val_size, batch_size', [(0.1, 100),
                                                  (0.5, 'all')])
def test_prep_data_dataloader(flow_model, data_dim, val_size, batch_size):
    """
    Test the data prep, make sure batch sizes and validation size
    produce the correct result.
    """
    n = 100
    x = np.random.randn(n, data_dim)

    train, val = flow_model.prep_data(
        x, val_size, batch_size, use_dataloader=True)
    train_batch = next(iter(train))[0]
    val_batch = next(iter(val))[0]
    if batch_size == 'all':
        batch_size = int(n * (1 - val_size))

    assert train.batch_size == batch_size
    assert list(val_batch.shape) == [int(val_size * n), data_dim]
    assert len(train) * train_batch.shape[0] + val_batch.shape[0] == n


@pytest.mark.parametrize('batch_size', [None, '10', True, False])
def test_incorrect_batch_size_type(flow_model, data_dim, batch_size):
    """Ensure the non-interger batch sizes do not work"""
    n = 1000
    x = np.random.randn(n, data_dim)
    with pytest.raises(RuntimeError) as excinfo:
        flow_model.prep_data(x, 0.1, batch_size)
    assert 'Unknown batch size' in str(excinfo.value)


@pytest.mark.parametrize('dataloader', [False, True])
def test_training(flow_model, data_dim, dataloader):
    """Test class until training"""
    x = np.random.randn(1000, data_dim)
    flow_model.use_dataloader = dataloader
    flow_model.train(x)
    assert flow_model.weights_file is not None


@pytest.mark.parametrize('key, value', [('annealing', True),
                                        ('noise_scale', 0.1),
                                        ('noise_scale', 'adaptive'),
                                        ('max_epochs', 51)])
def test_training_additional_config_args(flow_config, data_dim, tmpdir,
                                         key, value):
    """
    Test training with different config args
    """
    flow_config['model_config']['n_inputs'] = data_dim
    flow_config[key] = value

    output = str(tmpdir.mkdir('flowmodel'))
    flow_model = FlowModel(flow_config, output=output)

    assert getattr(flow_model, key) == value

    x = np.random.randn(100, data_dim)
    flow_model.train(x)


def test_early_optimiser_init(flow_model):
    """Ensure calling the opitmiser before the model raises an error"""
    with pytest.raises(RuntimeError) as excinfo:
        flow_model.get_optimiser()
    assert 'Cannot initialise optimiser' in str(excinfo.value)


@pytest.mark.parametrize('weights', [False, True])
@pytest.mark.parametrize('perms', [False, True])
def test_reset_model(flow_model, weights, perms):
    """Test resetting the model"""
    flow_model.initialise()
    flow_model.reset_model(weights=weights, permutations=perms)


def test_sample_and_log_prob_not_initialised(flow_model, data_dim):
    """
    Ensure user cannot call the method before the model initialise.
    """
    with pytest.raises(RuntimeError) as excinfo:
        flow_model.sample_and_log_prob()
    assert 'Model is not initialised' in str(excinfo.value)


@pytest.mark.parametrize('N', [1, 100])
def test_sample_and_log_prob_no_latent(flow_model, data_dim, N):
    """
    Test the basic use of sample and log prob and ensure correct output
    shapes.
    """
    flow_model.initialise()
    x, log_prob = flow_model.sample_and_log_prob(N=N)
    assert x.shape == (N, data_dim)
    assert log_prob.size == N


@pytest.mark.parametrize('N', [1, 100])
def test_sample_and_log_prob_with_latent(flow_model, data_dim, N):
    """
    Test the basic use of sample and log prob when samples from the
    latent space are provided
    """
    flow_model.initialise()
    z = np.random.randn(N, data_dim)
    x, log_prob = flow_model.sample_and_log_prob(z=z)
    assert x.shape == (N, data_dim)
    assert log_prob.size == N


@pytest.mark.parametrize('N', [1, 100])
def test_forward_and_log_prob(flow_model, data_dim, N):
    """Test the basic use of forward and log prob"""
    flow_model.initialise()
    x = np.random.randn(N, data_dim)
    z, log_prob = flow_model.forward_and_log_prob(x)
    assert z.shape == (N, data_dim)
    assert log_prob.size == N


def test_move_to_update_default(model):
    """Ensure the stored device is updated"""
    model.device = 'cuda'
    model.model = MagicMock()
    model.model.to = MagicMock()
    FlowModel.move_to(model, 'cpu', update_default=True)
    assert model.device == torch.device('cpu')
    model.model.to.assert_called_once()


@patch('torch.save')
def test_save_weights(mock_save, model):
    """Test saving the weights"""
    model.model = MagicMock()
    FlowModel.save_weights(model, 'model.pt')
    mock_save.assert_called_once()
    assert model.weights_file == 'model.pt'


def test_get_state(flow_model):
    """Make the object can be pickled"""
    pickle.dumps(flow_model)



#!/usr/bin/env python

# Example of using nessai for a Rosenbrock likelihood

import numpy as np

from nessai.flowsampler import FlowSampler
from nessai.model import Model
from nessai.utils import setup_logger
from nessai.livepoint import live_points_to_array

output = './outdir/rosenbrock/'
logger = setup_logger(output=output, log_level='WARNING')


class RosenbrockModel(Model):
    """Rosenbrock function defined in n dimensions on [-5, 5]^n.

    Based on the example in cpnest:\
        https://github.com/johnveitch/cpnest/blob/master/examples/rosenbrock.py
    """
    def __init__(self, dims):
        self.names = [f'x_{d}' for d in range(dims)]
        self.bounds = {n: [-5.0, 5.0] for n in self.names}

    def log_prior(self, x):
        """Uniform prior"""
        log_p = np.log(self.in_bounds(x))
        for n in self.names:
            log_p -= np.log(self.bounds[n][1] - self.bounds[n][0])
        return log_p

    def log_likelihood(self, x):
        """Log-likelihood"""
        x = live_points_to_array(x, self.names)[np.newaxis, :]
        return -(np.sum(
            100. * (x[:, 1:] - x[:, :-1] ** 2.) ** 2. + (1. - x[:, :-1]) ** 2.,
            axis=1)
        )


model = RosenbrockModel(2)

# Use a more complex flow compared to simpler examples.
flow_config = dict(
    batch_size=1000,
    max_epochs=200,
    patience=20,
    model_config=dict(n_blocks=4, n_neurons=4, n_layers=1)
)

# Configure the sampler.
# We use the logit rescaling since parts of the likelihood are close to the
# prior bounds
fs = FlowSampler(
    model,
    output=output,
    flow_config=flow_config,
    resume=False,
    seed=1234,
    nlive=2000,
    poolsize=2000,
    maximum_uninformed=4000,
    proposal_plots=False,
    reparameterisations={'logit': {'parameters': model.names}}
)

# And go!
fs.run()



# -*- coding: utf-8 -*-
"""
General utilities for nessai.
"""
from .distributions import (
    get_multivariate_normal,
    get_uniform_distribution
)
from .hist import auto_bins
from .indices import bonferroni_correction, compute_indices_ks_test
from .io import (
    NessaiJSONEncoder,
    is_jsonable,
    safe_file_dump,
    save_live_points
)
from .logging import setup_logger
from .rescaling import (
    configure_edge_detection,
    detect_edge,
    determine_rescaled_bounds,
    inverse_rescale_minus_one_to_one,
    inverse_rescale_zero_to_one,
    logit,
    rescale_minus_one_to_one,
    rescale_zero_to_one,
    rescaling_functions,
    sigmoid
)
from .sampling import (
    draw_gaussian,
    draw_nsphere,
    draw_surface_nsphere,
    draw_truncated_gaussian,
    draw_uniform
)
from .distance import compute_minimum_distances
from .structures import replace_in_list
from .threading import configure_threads


__all__ = [
    'BoxUniform',
    'MultivariateNormal',
    'NessaiJSONEncoder',
    'auto_bins',
    'bonferroni_correction',
    'compute_indices_ks_test',
    'compute_minimum_distances',
    'configure_edge_detection',
    'configure_threads',
    'detect_edge',
    'determine_rescaled_bounds',
    'distance', 'distributions',
    'draw_gaussian',
    'draw_nsphere',
    'draw_surface_nsphere',
    'draw_truncated_gaussian',
    'draw_uniform',
    'get_multivariate_normal',
    'get_uniform_distribution',
    'hist',
    'indices',
    'inverse_rescale_minus_one_to_one',
    'inverse_rescale_zero_to_one',
    'is_jsonable',
    'logit',
    'replace_in_list',
    'rescale_minus_one_to_one',
    'rescale_zero_to_one',
    'rescaling',
    'rescaling_functions',
    'safe_file_dump',
    'sampling',
    'save_live_points',
    'setup_logger',
    'sigmoid',
    'spatial',
    'structures',
    'threading'
]



# -*- coding: utf-8 -*-
"""
Test io utilities.
"""
import os
import json
import numpy as np
import pickle
import pytest
from unittest.mock import call, create_autospec, mock_open, patch

from nessai.livepoint import numpy_array_to_live_points
from nessai.utils.io import (
    NessaiJSONEncoder,
    is_jsonable,
    safe_file_dump,
    save_live_points
)


def test_is_jsonable_true():
    """Assert True is return is json.dumps does not raise an error"""
    assert is_jsonable({'x': 2})


@pytest.mark.parametrize('cls', [TypeError, OverflowError])
def test_is_jsonable_false(cls, ):
    """Assert True is return is json.dumps does not raise an error"""
    with patch('json.dumps', side_effect=cls()) as mock:
        assert not is_jsonable({'x': 2})
    mock.assert_called_once_with({'x': 2})


@pytest.mark.parametrize(
    'input, expected',
    [(np.int32(2), 2), (np.float64(2), 2.0), (np.array([1, 2]), [1, 2])]
)
def test_JSON_encoder_numpy(input, expected):
    """Test the JSON encoder with numpy inputs"""
    e = create_autospec(NessaiJSONEncoder)
    output = NessaiJSONEncoder.default(e, input)
    assert output == expected


def test_JSON_encoder_object():
    """Test the JSON with an object that is JSONable"""
    e = create_autospec(NessaiJSONEncoder)
    input = object()
    with patch('nessai.utils.io.is_jsonable', return_value=False) as m:
        output = NessaiJSONEncoder.default(e, input)
    m.assert_called_once_with(input)
    assert 'object object at' in output


def test_JSON_encoder_other():
    """Test the JSON with an object that is JSONable"""
    e = create_autospec(NessaiJSONEncoder)
    input = 'Hello'
    with patch('json.JSONEncoder.default', return_value='Hello') as m:
        output = NessaiJSONEncoder.default(e, input)
    m.assert_called_once_with(input)
    assert output == 'Hello'


def test_safe_file_dump():
    """Test safe file dump."""
    m = mock_open()
    data = np.array([1, 2])
    with patch('builtins.open', m) as mo,\
         patch('pickle.dump') as md,\
         patch('shutil.move') as msm:
        safe_file_dump(data, 'test.pkl', pickle, save_existing=False)
    mo.assert_called_once_with('test.pkl.temp', 'wb')
    md.call_args_list[0][0] == data
    msm.assert_called_once_with('test.pkl.temp', 'test.pkl')


def test_safe_file_dump_save_existing():
    """Test safe file dump."""
    import pickle
    m = mock_open()
    data = np.array([1, 2])
    with patch('os.path.exists', return_value=True) as mpe,\
         patch('builtins.open', m) as mo,\
         patch('pickle.dump') as md,\
         patch('shutil.move') as msm:
        safe_file_dump(data, 'test.pkl', pickle, save_existing=True)
    mpe.assert_called_once_with('test.pkl')
    mo.assert_called_once_with('test.pkl.temp', 'wb')
    md.call_args_list[0][0] == data
    msm.assert_has_calls([
        call('test.pkl', 'test.pkl.old'), call('test.pkl.temp', 'test.pkl'),
    ])


def test_safe_file_dump_integration(tmp_path):
    """Integration test for safe file dump"""
    path = tmp_path
    f = path / 'test.pkl'
    f.write_text('a')
    data = 'b'
    safe_file_dump(data, str(path) + '/test.pkl', pickle, save_existing=True)
    assert os.path.exists(str(path) + '/test.pkl')
    assert os.path.exists(str(path) + '/test.pkl.old')


def test_save_live_points(tmp_path):
    """Test the function for saving live points"""
    d = {'x': [1, 2], 'y': [3, 4], 'logP': [0, 0], 'logL': [0, 0]}
    live_points = numpy_array_to_live_points(
        np.array([[1, 3], [2, 4]]), ['x', 'y']
    )
    filename = str(tmp_path) + 'test.json'
    save_live_points(live_points, filename)

    with open(filename, 'r') as fp:
        d_out = json.load(fp)

    assert d_out == d



# -*- coding: utf-8 -*-
"""Test the GW flow proposal method"""
import pytest
from unittest.mock import create_autospec, MagicMock, patch

from nessai.gw.proposal import GWFlowProposal, AugmentedGWFlowProposal


@pytest.fixture
def proposal():
    return create_autospec(GWFlowProposal)


@pytest.fixture
def augmented_proposal():
    return create_autospec(AugmentedGWFlowProposal)


def test_get_reparameterisation(proposal):
    """Test to make sure the correct version of get reparameterisation is
    called.
    """
    with patch('nessai.gw.proposal.get_gw_reparameterisation',
               return_value='reparam') as mock:
        out = GWFlowProposal.get_reparameterisation(proposal, 'default')
    assert out == 'reparam'
    mock.assert_called_once_with('default')


def test_add_default_reparameterisation(proposal):
    """Test the method `add_default_reparameterisation`"""
    proposal.aliases = GWFlowProposal.aliases
    proposal.names = ['chirp_mass', 'theta_jn']
    proposal._reparameterisation = MagicMock()
    proposal._reparameterisation.parameters = ['chirp_mass']
    proposal.model = MagicMock()
    proposal.model.bounds = \
        {'chirp_mass': [10.0, 20.0], 'theta_jn': [0.0, 3.0]}

    reparam = MagicMock()
    with patch('nessai.gw.proposal.get_gw_reparameterisation',
               return_value=(reparam, {})) as mock_get:
        GWFlowProposal.add_default_reparameterisations(proposal)

    mock_get.assert_called_once_with('angle-sine')
    reparam.assert_called_once_with(parameters=['theta_jn'],
                                    prior_bounds={'theta_jn': [0.0, 3.0]})


def test_augmented_get_reparameterisation(augmented_proposal):
    """Test to make sure the correct version of get reparameterisation is
    called.
    """
    with patch('nessai.gw.proposal.get_gw_reparameterisation',
               return_value='reparam') as mock:
        out = AugmentedGWFlowProposal.get_reparameterisation(
            augmented_proposal, 'default')
    assert out == 'reparam'
    mock.assert_called_once_with('default')


def test_augmented_reparameterisation_prior(augmented_proposal):
    """Test to make sure the correct components of the log prior are being
    called.

    These are:
    - agumented_prior
    - FlowProposal.log_prior
    """
    augmented_proposal.augmented_prior = MagicMock()
    augmented_proposal._reparameterisation = MagicMock()
    augmented_proposal.model = MagicMock()

    with patch('nessai.proposal.flowproposal.FlowProposal.log_prior') as mock:
        AugmentedGWFlowProposal.log_prior(augmented_proposal, 1)

    augmented_proposal.augmented_prior.assert_called_once_with(1)
    mock.assert_called_once_with(1)


def test_augmented_reparameterisation_prime_prior(augmented_proposal):
    """Test to make sure the correct components of the primed_log prior are
    being called.

    These are:
    - agumented_prior
    - FlowProposal.x_prime_log_prior
    """
    augmented_proposal.augmented_prior = MagicMock()
    augmented_proposal._reparameterisation = MagicMock()

    with patch('nessai.proposal.flowproposal.FlowProposal.'
               'x_prime_log_prior') as mock:
        AugmentedGWFlowProposal.x_prime_log_prior(augmented_proposal, 1)

    augmented_proposal.augmented_prior.assert_called_once_with(1)
    mock.assert_called_once_with(1)



# -*- coding: utf-8 -*-
"""
Test the object that handles the nested sampling evidence and prior volumes.
"""
import numpy as np
import pytest

from nessai.evidence import (
    _NSIntegralState,
    logsubexp,
)


@pytest.fixture()
def nlive():
    return 100


def test_logsubexp_negative():
    """
    Test behaviour of logsubexp for x < y
    """
    with pytest.raises(Exception):
        logsubexp(1, 2)


def test_increment(nlive):
    """Test the basic functionality of incrementing the evidence estimate"""
    state = _NSIntegralState(nlive)
    state.increment(-10)

    assert state.logw == (-1 / nlive)
    assert state.logZ != -np.inf
    np.testing.assert_equal(state.logLs, [-np.inf, -10])


def test_finalise(nlive):
    """Test the check that finalise returns an improved logZ"""
    state = _NSIntegralState(nlive)
    state.increment(-10)
    pre = state.logZ
    state.finalise()
    assert state.logZ != -np.inf
    assert pre != state.logZ


def test_info(nlive):
    """Test to check the information increases as expected"""
    state = _NSIntegralState(nlive)
    state.increment(-10)
    assert state.info == [0.]
    state.increment(-5)
    assert state.info[1] > 0


def test_track_gradients(nlive):
    """
    Test to make sure gradients are not computed when tracking is disabled
    """
    state = _NSIntegralState(nlive, track_gradients=False)
    state.increment(-10)
    state.increment(-5)
    assert len(state.gradients) == 1


def test_variable_nlive(nlive):
    """
    Test to make sure that the using a different nlive changes the update
    values.
    """
    state = _NSIntegralState(nlive)
    state.increment(-10, nlive=50)
    assert state.logw == (-1 / 50)


def test_plot(nlive):
    """Test the plotting function"""
    state = _NSIntegralState(nlive)
    state.increment(-10)
    state.increment(-5)
    fig = state.plot()
    assert fig is not None


def test_plot_w_filename(nlive, tmpdir):
    """Test the plotting function with a filename specified"""
    filename = str(tmpdir.mkdir('test'))
    state = _NSIntegralState(nlive)
    state.increment(-10)
    state.increment(-5)
    fig = state.plot(filename=filename)
    assert fig is None



# -*- coding: utf-8 -*-
"""Test the flow utilities"""
import logging
import numpy as np
import pytest
import torch
import torch.nn.functional as F
from unittest.mock import MagicMock, create_autospec, patch

from nessai.flows.utils import (
    configure_model,
    silu,
    reset_weights,
    reset_permutations,
    MLP
)


@pytest.fixture
def config():
    """Minimal config needed for configure_model to work"""
    return dict(
        n_inputs=2,
        n_neurons=4,
        n_blocks=2,
        n_layers=1,
        ftype='realnvp'
    )


def test_silu():
    """Test the silu activation"""
    from scipy.special import expit
    x = torch.randn(100)
    y = silu(x)
    expected = x.numpy() * expit(x.numpy())
    np.testing.assert_array_almost_equal(y, expected)


def test_reset_weights_with_reset_parameters():
    """Test the reset weights function for module with ``reset_parameters``"""
    module = MagicMock()
    module.reset_parameters = MagicMock()
    reset_weights(module)
    module.reset_parameters.assert_called_once()


def test_reset_weights_batch_norm():
    """Test the reset weights function for an instance of batch norm"""
    from nflows.transforms.normalization import BatchNorm
    x = torch.randn(20, 2)
    module = BatchNorm(2, eps=0.1)
    module.train()
    module.forward(x)

    reset_weights(module)

    constant = np.log(np.exp(1 - 0.1) - 1)
    assert (module.unconstrained_weight.data == constant).all()
    assert (module.bias.data == 0).all()
    assert (module.running_mean == 0).all()
    assert (module.running_var == 1).all()


def test_reset_weights_other_module(caplog):
    """Test reset weights on a module that cannot be reset."""
    caplog.set_level(logging.WARNING)
    module = object
    reset_weights(module)
    assert 'Could not reset' in caplog.text


def test_weight_reset_permutation():
    """Test to make sure random permutation is reset correctly"""
    from nflows.transforms.permutations import RandomPermutation
    x = torch.arange(10).reshape(1, -1)
    m = RandomPermutation(features=10)
    y_init, _ = m(x)
    p = m._permutation.numpy()
    m.apply(reset_permutations)
    y_reset, _ = m(x)
    assert not (p == m._permutation.numpy()).all()
    assert not (y_init.numpy() == y_reset.numpy()).all()


def test_mlp_forward():
    """Test the MLP implementation of forward"""
    mlp = create_autospec(MLP)
    x = torch.tensor(1)
    y = torch.tensor(2)
    with patch('nflows.nn.nets.MLP.forward',
               return_value=y) as parent:
        out = MLP.forward(mlp, x, context=None)
    parent.assert_called_once_with(x)
    assert out == y


def test_mlp_forward_context():
    """Assert an error is raised in the context is not None"""
    mlp = create_autospec(MLP)
    x = torch.tensor(1)
    with pytest.raises(NotImplementedError) as excinfo:
        MLP.forward(mlp, x, context=x)
    assert 'MLP with conditional inputs is not implemented.' in \
        str(excinfo.value)


def test_configure_model_basic(config):
    """Test configure model with the most basic config."""
    config['kwargs'] = dict(num_bins=2)
    with patch('nessai.flows.utils.RealNVP') as mock_flow:
        configure_model(config)

    mock_flow.assert_called_with(
        config['n_inputs'],
        config['n_neurons'],
        config['n_blocks'],
        config['n_layers'],
        num_bins=2
    )


@pytest.mark.parametrize(
    'flow_inputs',
    [
        {'ftype': 'realnvp', 'expected': 'RealNVP'},
        {'ftype': 'frealnvp', 'expected': 'RealNVP'},
        {'ftype': 'spline', 'expected': 'NeuralSplineFlow'},
        {'ftype': 'nsf', 'expected': 'NeuralSplineFlow'},
        {'ftype': 'maf', 'expected': 'MaskedAutoregressiveFlow'}
    ]
)
def test_configure_model_flows(config, flow_inputs):
    """Test the different flows."""
    config['ftype'] = flow_inputs['ftype']
    with patch(f"nessai.flows.utils.{flow_inputs['expected']}") as mock_flow:
        model, _ = configure_model(config)
    mock_flow.assert_called_with(
        config['n_inputs'],
        config['n_neurons'],
        config['n_blocks'],
        config['n_layers'],
    )


def test_configure_model_flow_class(config):
    """Test using a custom class of flow."""
    class TestFlow:
        def __init__(self, n_inputs, n_neurons, n_blocks, n_layers):
            self.n_inputs = n_inputs
            self.n_neurons = n_neurons
            self.n_blocks = n_blocks
            self.n_layers = n_layers

        def to(self, input):
            pass

    config['flow'] = TestFlow
    model, _ = configure_model(config)
    assert isinstance(model, TestFlow)
    assert model.n_inputs == config['n_inputs']
    assert model.n_neurons == config['n_neurons']
    assert model.n_blocks == config['n_blocks']
    assert model.n_layers == config['n_layers']


def test_configure_model_device_cuda(config):
    config['device_tag'] = 'cuda'
    expected_device = torch.device('cuda')
    mock_model = MagicMock()
    with patch('nessai.flows.utils.RealNVP',
               return_value=mock_model) as mock_flow:
        model, device = configure_model(config)

    mock_flow.assert_called_with(
        config['n_inputs'],
        config['n_neurons'],
        config['n_blocks'],
        config['n_layers'],
    )

    mock_model.to.assert_called_once_with(expected_device)
    assert model.device == expected_device
    assert device == expected_device


@pytest.mark.parametrize(
    'act',
    [
        {'act': 'relu', 'expected': F.relu},
        {'act': 'tanh', 'expected': F.tanh},
        {'act': 'silu', 'expected': silu},
        {'act': 'swish', 'expected': silu}
    ]
)
def test_configure_model_activation_functions(config, act):
    """Test the different activation functions."""
    config['kwargs'] = dict(activation=act['act'])

    with patch('nessai.flows.utils.RealNVP') as mock_flow:
        configure_model(config)

    mock_flow.assert_called_with(
        config['n_inputs'],
        config['n_neurons'],
        config['n_blocks'],
        config['n_layers'],
        activation=act['expected']
    )


def test_configure_model_ftype_error(config):
    """Assert unknown types of flow raise an error."""
    config.pop('ftype')
    with pytest.raises(RuntimeError) as excinfo:
        configure_model(config)
    assert "Must specify either 'flow' or 'ftype'." in str(excinfo.value)


def test_configure_model_input_type_error(config):
    """Assert incorrect type for n_inputs raises an error."""
    config['n_inputs'] = '10'
    with pytest.raises(TypeError) as excinfo:
        configure_model(config)
    assert 'Number of inputs (n_inputs) must be an int' in str(excinfo.value)


def test_configure_model_unknown_activation(config):
    """Assert unknown activation functions raise an error"""
    config['kwargs'] = dict(activation='test')
    with pytest.raises(RuntimeError) as excinfo:
        configure_model(config)
    assert "Unknown activation function: 'test'" in str(excinfo.value)



# -*- coding: utf-8 -*-
"""
Utilities related to drawing samples.
"""
import numpy as np
from scipy import stats


def draw_surface_nsphere(dims, r=1, N=1000):
    """
    Draw N points uniformly from  n-1 sphere of radius r using Marsaglia's
    algorithm. E.g for 3 dimensions returns points on a 'regular' sphere.

    See Marsaglia (1972)

    Parameters
    ----------
    dims : int
        Dimension of the n-sphere
    r : float, optional
        Radius of the n-sphere, if specified it is used to rescale the samples
    N : int, optional
        Number of samples to draw

    Returns
    -------
    ndarray
        Array of samples with shape (N, dims)
    """
    x = np.random.randn(N, dims)
    R = np.sqrt(np.sum(x ** 2., axis=1))[:, np.newaxis]
    z = x / R
    return r * z


def draw_nsphere(dims, r=1, N=1000, fuzz=1.0):
    """
    Draw N points uniformly within an n-sphere of radius r

    Parameters
    ----------
    dims : int
        Dimension of the n-sphere
    r : float, optional
        Radius of the n-ball
    N : int, optional
        Number of samples to draw
    fuzz : float, optional
        Fuzz factor by which to increase the radius of the n-ball

    Returns
    -------
    ndarray
        Array of samples with shape (N, dims)
    """
    x = draw_surface_nsphere(dims, r=1, N=N)
    R = np.random.uniform(0, 1, (N, 1))
    z = R ** (1 / dims) * x
    return fuzz * r * z


def draw_uniform(dims, r=(1,), N=1000, fuzz=1.0):
    """
    Draw from a uniform distribution on [0, 1].

    Deals with extra input parameters used by other draw functions

    Parameters
    ----------
    dims : int
        Dimension of the n-sphere
    r : float, optional
        Radius of the n-ball. (Ignored by this function)
    N : int, ignored
        Number of samples to draw
    fuzz : float, ignored
        Fuzz factor by which to increase the radius of the n-ball. (Ignored by
        this function)

    Returns
    -------
    ndarraay
        Array of samples with shape (N, dims)
    """
    return np.random.uniform(0, 1, (N, dims))


def draw_gaussian(dims, r=1, N=1000, fuzz=1.0):
    """
    Wrapper for numpy.random.randn that deals with extra input parameters
    r and fuzz

    Parameters
    ----------
    dims : int
        Dimension of the n-sphere
    r : float, optional
        Radius of the n-ball
    N : int, ignored
        Number of samples to draw
    fuzz : float, ignored
        Fuzz factor by which to increase the radius of the n-ball

    Returns
    -------
    ndarray
        Array of samples with shape (N, dims)
    """
    return np.random.randn(N, dims)


def draw_truncated_gaussian(dims, r, N=1000, fuzz=1.0, var=1):
    """
    Draw N points from a truncated gaussian with a given a radius

    Parameters
    ----------
    dims : int
        Dimension of the n-sphere
    r : float
        Radius of the truncated Gaussian
    N : int, ignored
        Number of samples to draw
    fuzz : float, ignored
        Fuzz factor by which to increase the radius of the truncated Gaussian

    Returns
    -------
    ndarray
        Array of samples with shape (N, dims)
    """
    sigma = np.sqrt(var)
    r *= fuzz
    u_max = stats.chi.cdf(r / sigma, df=dims)
    u = np.random.uniform(0, u_max, N)
    p = sigma * stats.chi.ppf(u, df=dims)
    x = np.random.randn(p.size, dims)
    points = (p * x.T / np.sqrt(np.sum(x**2., axis=1))).T
    return points



#!/usr/bin/env python
# Example of using nessai with the eggbox function. Nessai is not well
# suited to this sorted problem. The augmented proposal method is used to
# to improve the results but ultimately likelihoods that this multi-modal
# are a challenge.

import numpy as np

from nessai.flowsampler import FlowSampler
from nessai.model import Model
from nessai.utils import setup_logger


output = './outdir/eggbox/'
logger = setup_logger(output=output, log_level='INFO')


class EggboxModel(Model):
    """
    Eggbox problem from https://arxiv.org/pdf/0809.3437v1.pdf.

    Based on the example in cpnest: \
        https://github.com/johnveitch/cpnest/blob/master/examples/eggbox.py
    """

    def __init__(self, dims):
        self.names = [str(i) for i in range(dims)]
        self.bounds = {n: [0, 10 * np.pi] for n in self.names}

    def log_prior(self, x):
        log_p = 0.
        for n in self.names:
            log_p += (np.log((x[n] >= self.bounds[n][0])
                             & (x[n] <= self.bounds[n][1]))
                      - np.log(self.bounds[n][1] - self.bounds[n][0]))
        return log_p

    def log_likelihood(self, x):
        log_l = 1.0
        for n in self.names:
            log_l *= np.cos(x[n] / 2.)
        return (log_l + 2.0) ** 5.0


flow_config = dict(
        batch_size=1000,
        max_epochs=200,
        patience=50,
        model_config=dict(n_blocks=4, n_neurons=8, n_layers=2, ftype='realnvp',
                          kwargs=dict(batch_norm_between_layers=False,
                                      linear_transform='lu'))
        )


fs = FlowSampler(EggboxModel(2), output=output, flow_config=flow_config,
                 resume=False, seed=1234, flow_class='AugmentedFlowProposal',
                 augment_features=4, expansion_fraction=2.0, nlive=2000,
                 poolsize=2000, maximum_uninformed=np.inf,
                 update_poolsize=True, max_threads=2, n_pool=1)

fs.run()



#!/usr/bin/env python

# Example of using AugmentedFlowProposal for a multimodel model

import numpy as np
from scipy.stats import norm

from nessai.flowsampler import FlowSampler
from nessai.model import Model
from nessai.utils import setup_logger

output = './outdir/augmented_example/'
logger = setup_logger(output=output, log_level='INFO')

# Define the model


class MultimodalModel(Model):
    """
    A multimodal gaussian model.
    """
    def __init__(self, dims=2):
        self.names = [str(i) for i in range(dims)]
        self.bounds = {i: [-10, 10] for i in self.names}

    def log_prior(self, x):
        """
        Returns log of prior given a live point assuming uniform
        priors on each parameter.
        """
        log_p = 0.
        # Iterate through each parameter (x and y)
        # since the live points are a structured array we can
        # get each value using just the name
        for n in self.names:
            log_p += (np.log((x[n] >= self.bounds[n][0])
                             & (x[n] <= self.bounds[n][1]))
                      - np.log(self.bounds[n][1] - self.bounds[n][0]))
        return log_p

    def log_likelihood(self, x):
        """
        Returns log likelihood of given live point.
        """
        log_l = 0
        # Use a Gaussian logpdf and iterate through the parameters
        for pn in self.names:
            log_l += np.log(norm(-5).pdf(x[pn]) + norm(5).pdf(x[pn]))
        return log_l


flow_config = dict(
        batch_size=1000,
        max_epochs=200,
        patience=50,
        model_config=dict(n_blocks=4, n_neurons=4, n_layers=2, ftype='realnvp',
                          kwargs=dict(batch_norm_between_layers=False,
                                      linear_transform='lu'))
        )

# We `augment_dims` which help the sampler bridge disconnected regions of
# probability. There additional parameters are Gaussian and included in the
# parameters transformed by the flow.
fp = FlowSampler(MultimodalModel(4), output=output, flow_config=flow_config,
                 resume=False, seed=1234, augment_dims=2, nlive=2000,
                 generate_augment='gaussian', expansion_fraction=2.0,
                 flow_class='AugmentedFlowProposal', maximum_uninformed=4000,
                 poolsize=2000)

# And go!
fp.run()



# -*- coding: utf-8 -*-
"""
Tests for `nessai.model`
"""
import numpy as np
import pytest
from scipy.stats import norm
from unittest.mock import MagicMock, create_autospec

from nessai.livepoint import numpy_array_to_live_points
from nessai.model import Model


class EmptyModel(Model):

    def log_prior(self, x):
        return None

    def log_likelihood(self, x):
        return None


@pytest.fixture()
def integration_model():
    class TestModel(Model):

        def __init__(self):
            self.bounds = {'x': [-5, 5], 'y': [-5, 5]}
            self.names = ['x', 'y']

        def log_prior(self, x):
            log_p = np.log(self.in_bounds(x))
            for n in self.names:
                log_p -= np.log(self.bounds[n][1] - self.bounds[n][0])
            return log_p

        def log_likelihood(self, x):
            log_l = 0
            for pn in self.names:
                log_l += norm.logpdf(x[pn])
            return log_l

    return TestModel()


@pytest.fixture
def model():
    return create_autospec(Model)


def test_dims_no_names(model):
    """Test the behaviour dims when names is empty"""
    assert Model.dims.__get__(model) is None


def test_dims(model):
    """Ensure dims are correct"""
    model.names = ['x', 'y']
    assert Model.dims.__get__(model) == 2


def test_lower_bounds(model):
    """Check the lower bounds are correctly set"""
    model.bounds = {'x': [-1, 1], 'y': [-1, 1]}
    model._lower = None
    assert (Model.lower_bounds.__get__(model) == [-1, -1]).all()


def test_upper_bounds(model):
    """Check the upper bounds are correctly set"""
    model.bounds = {'x': [-1, 1], 'y': [-1, 1]}
    model._upper = None
    assert (Model.upper_bounds.__get__(model) == [1, 1]).all()


def test_in_bounds(model):
    """Test the `in_bounds` method.

    Tests both finite and infinite prior bounds.
    """
    x = numpy_array_to_live_points(np.array([[0.5, 1], [2, 1]]), ['x', 'y'])
    model.names = ['x', 'y']
    model.bounds = {'x': [0, 1], 'y': [-np.inf, np.inf]}
    val = Model.in_bounds(model, x)
    np.testing.assert_array_equal(val, np.array([True, False]))


def test_parameter_in_bounds(model):
    """Test parameter in bounds method."""
    x = np.array([0, 0.5, 1, 3])
    model.names = ['x', 'y']
    model.bounds = {'x': [0, 1], 'y': [0, 4]}
    val = Model.parameter_in_bounds(model, x, 'x')
    np.testing.assert_array_equal(val, np.array([True, True, True, False]))


def test_sample_parameter(model):
    """Assert an error is raised."""
    with pytest.raises(NotImplementedError) as excinfo:
        Model.sample_parameter(model, 'x', n=2)
    assert 'User must implement this method!' in str(excinfo.value)


def test_new_point_single(model):
    """Test the new point when asking for 1 point"""
    model._single_new_point = MagicMock()
    Model.new_point(model, N=1)
    model._single_new_point.assert_called_once()


def test_single_new_point(model):
    """Test the method that draw one point within the prior bounds"""
    model.names = ['x', 'y']
    model.bounds = {'x': [-1, 1], 'y': [-2, 2]}
    model.lower_bounds = np.array([-1, -2])
    model.upper_bounds = np.array([1, 2])
    model.log_prior = MagicMock(return_value=0)
    model.dims = 2
    x = Model._single_new_point(model)
    assert ((x['x'] >= -1) & (x['x'] <= 1))
    assert ((x['y'] >= -2) & (x['y'] <= 2))


def test_new_point_multiple(model):
    """Test the new point when asking for multiple points"""
    model._multiple_new_points = MagicMock()
    Model.new_point(model, N=10)
    model._multiple_new_points.assert_called_once_with(10)


def test_multiple_new_points(model):
    """Test the method that draws multiple points within the prior bounds"""
    n = 10
    model.names = ['x', 'y']
    model.bounds = {'x': [-1, 1], 'y': [-2, 2]}
    model.lower_bounds = np.array([-1, -2])
    model.upper_bounds = np.array([1, 2])
    model.log_prior = MagicMock(return_value=np.zeros(10))
    model.dims = 2
    x = Model._multiple_new_points(model, N=n)
    assert x.size == n
    assert ((x['x'] >= -1) & (x['x'] <= 1)).all()
    assert ((x['y'] >= -2) & (x['y'] <= 2)).all()


def test_new_point_log_prob(model):
    """Test the log prob for new points.

    Should be zero.
    """
    x = numpy_array_to_live_points(np.random.randn(2, 1), ['x'])
    log_prob = Model.new_point_log_prob(model, x)
    assert log_prob.size == 2
    assert (log_prob == 0).all()


@pytest.mark.integration_test
def test_new_point_integration(integration_model):
    """
    Test the default method for generating a new point with the bounds.

    Uses the model defined in `conftest.py` with bounds [-5, 5] for
    x and y.
    """
    new_point = integration_model.new_point()
    log_q = integration_model.new_point_log_prob(new_point)
    assert (new_point['x'] < 5) & (new_point['y'] > -5)
    assert (new_point['y'] < 5) & (new_point['y'] > -5)
    assert log_q == 0


@pytest.mark.integration_test
def test_new_point_multiple_integration(integration_model):
    """
    Test drawing multiple new points from the model

    Uses the model defined in `conftest.py` with bounds [-5, 5] for
    x and y.
    """
    new_points = integration_model.new_point(N=100)
    log_q = integration_model.new_point_log_prob(new_points)
    assert new_points.size == 100
    assert all(np.isfinite(new_points['logP']))
    assert all(new_points['x'] < 5) & all(new_points['x'] > -5)
    assert all(new_points['y'] < 5) & all(new_points['y'] > -5)
    assert (log_q == 0).all()


def test_likelihood_evaluations(model):
    """
    Test `evaluate_log_likelihood` and ensure the counter increases.
    """
    x = 1
    model.likelihood_evaluations = 0
    model.log_likelihood = MagicMock(return_value=2)
    log_l = Model.evaluate_log_likelihood(model, x)

    model.log_likelihood.assert_called_once_with(x)
    assert log_l == 2
    assert model.likelihood_evaluations == 1


def test_log_prior(model):
    """Verify the log prior does nothing by default"""
    assert Model.log_prior(model, 1) is None


def test_log_likelihood(model):
    """Verify the log likelihood does nothing by default"""
    assert Model.log_likelihood(model, 1) is None


def test_missing_log_prior():
    """
    Test to ensure a model cannot be instantiated without a log-prior method.
    """
    class TestModel(Model):

        def __init__(self):
            self.bounds = {'x': [-5, 5], 'y': [-5, 5]}
            self.names = ['x', 'y']

        def log_likelihood(self, x):
            return x

    with pytest.raises(TypeError) as excinfo:
        TestModel()
    assert "Can't instantiate abstract class TestModel" in str(excinfo.value)


def test_missing_log_likelihood():
    """
    Test to ensure a model cannot be instantiated without a log-likelihood
    method.
    """
    class TestModel(Model):

        def __init__(self):
            self.bounds = {'x': [-5, 5], 'y': [-5, 5]}
            self.names = ['x', 'y']

        def log_prior(self, x):
            return 0.

    with pytest.raises(TypeError) as excinfo:
        TestModel()
    assert "Can't instantiate abstract class TestModel" in str(excinfo.value)


def test_verify_new_point():
    """
    Test `Model.verify_model` and ensure a model with an ill-defined
    prior function raises the correct error
    """
    class TestModel(EmptyModel):

        def __init__(self):
            self.bounds = {'x': [-5, 5], 'y': [-5, 5]}
            self.names = ['x', 'y']

        def log_prior(self, x):
            return -np.inf

    model = TestModel()

    with pytest.raises(RuntimeError) as excinfo:
        model.verify_model()

    assert 'valid point' in str(excinfo.value)


@pytest.mark.parametrize("log_p", [np.inf, -np.inf])
def test_verify_log_prior_finite(log_p):
    """
    Test `Model.verify_model` and ensure a model with a log-prior that
    only returns inf function raises the correct error
    """
    class TestModel(EmptyModel):

        def __init__(self):
            self.bounds = {'x': [-5, 5], 'y': [-5, 5]}
            self.names = ['x', 'y']

        def log_prior(self, x):
            return log_p

    model = TestModel()

    with pytest.raises(RuntimeError):
        model.verify_model()


def test_verify_log_prior_none():
    """
    Test `Model.verify_model` and ensure a model with a log-prior that
    only returns None raises an error.
    """
    class TestModel(EmptyModel):

        def __init__(self):
            self.bounds = {'x': [-5, 5], 'y': [-5, 5]}
            self.names = ['x', 'y']

        def log_prior(self, x):
            return None

    model = TestModel()

    with pytest.raises(RuntimeError) as excinfo:
        model.verify_model()

    assert 'Log-prior' in str(excinfo.value)


def test_verify_log_likelihood_none():
    """
    Test `Model.verify_model` and ensure a model with a log-likelihood that
    only returns None raises an error.
    """
    class TestModel(EmptyModel):

        def __init__(self):
            self.bounds = {'x': [-5, 5], 'y': [-5, 5]}
            self.names = ['x', 'y']

        def log_prior(self, x):
            return 0

        def log_likelihood(self, x):
            return None

    model = TestModel()

    with pytest.raises(RuntimeError) as excinfo:
        model.verify_model()

    assert 'Log-likelihood' in str(excinfo.value)


def test_verify_no_names():
    """
    Test `Model.verify_model` and ensure a model without names
    function raises the correct error
    """
    class TestModel(EmptyModel):

        def __init__(self):
            self.names = []
            self.bounds = {'x': [-5, 5], 'y': [-5, 5]}

    model = TestModel()

    with pytest.raises(ValueError):
        model.verify_model()


def test_verify_no_bounds():
    """
    Test `Model.verify_model` and ensure a model without bounds
    function raises the correct error
    """
    class TestModel(EmptyModel):

        def __init__(self):
            self.names = ['x', 'y']
            self.bounds = {}

    model = TestModel()

    with pytest.raises(ValueError) as excinfo:
        model.verify_model()

    assert 'Bounds are not set' in str(excinfo.value)


def test_verify_incomplete_bounds():
    """
    Test `Model.verify_model` and ensure a model without bounds
    function raises the correct error
    """
    class TestModel(EmptyModel):

        def __init__(self):
            self.names = ['x', 'y']
            self.bounds = {'x': [-5, 5]}

    model = TestModel()

    with pytest.raises(RuntimeError):
        model.verify_model()


def test_unbounded_priors_wo_new_point():
    """Test `Model.verify_model` with unbounded priors"""

    class TestModel(Model):

        def __init__(self):
            self.names = ['x', 'y']
            self.bounds = {'x': [-5, 5], 'y': [-np.inf, np.inf]}

        def log_prior(self, x):
            return -np.log(10) * np.ones(x.size) + norm.logpdf(x['y'])

        def log_likelihood(self, x):
            return np.ones(x.size)

    model = TestModel()
    with pytest.raises(RuntimeError) as excinfo:
        model.verify_model()

    assert 'Could not draw a new point' in str(excinfo.value)


def test_unbounded_priors_w_new_point():
    """Test `Model.verify_model` with unbounded priors"""

    class TestModel(Model):

        def __init__(self):
            self.names = ['x', 'y']
            self.bounds = {'x': [-5, 5], 'y': [-np.inf, np.inf]}

        def new_point(self, N=1):
            return numpy_array_to_live_points(
                    np.concatenate([np.random.uniform(-5, 5, (N, 1)),
                                    norm.rvs(size=(N, 1))], axis=1),
                    self.names)

        def log_prior(self, x):
            return -np.log(10) * np.ones(x.size) + norm.logpdf(x['y'])

        def log_likelihood(self, x):
            return np.ones(x.size)

    model = TestModel()
    model.verify_model()



# -*- coding: utf-8 -*-
"""Test general configuration functions"""
import pytest

from nessai.proposal import FlowProposal
from nessai import utils


def test_config_drawsize_none(proposal):
    """Test the popluation configuration with no drawsize given"""
    FlowProposal.configure_population(proposal, 2000, None, True, 10, 1.0, 0.0,
                                      'gaussian')
    assert proposal.drawsize == 2000


def test_config_poolsize_none(proposal):
    """
    Test the popluation configuration raises an error if poolsize is None.
    """
    with pytest.raises(RuntimeError) as excinfo:
        FlowProposal.configure_population(
            proposal, None, None, True, 10, 1.0, 0.0, 'gaussian')

    assert 'poolsize' in str(excinfo.value)


@pytest.mark.parametrize('fixed_radius', [False, 5.0, 1])
def test_config_fixed_radius(proposal, fixed_radius):
    """Test the configuration for a fixed radius"""
    FlowProposal.configure_fixed_radius(proposal, fixed_radius)
    assert proposal.fixed_radius == fixed_radius


def test_config_fixed_radius_not_float(proposal):
    """
    Test the fixed radius is disabled when the radius cannot be converted to
    a float.
    """
    FlowProposal.configure_fixed_radius(proposal, 'four')
    assert proposal.fixed_radius is False


def test_min_radius_no_max(proposal):
    """Test configuration of min radius and no max radius"""
    FlowProposal.configure_min_max_radius(proposal, 5.0, False)
    assert proposal.min_radius == 5.0
    assert proposal.max_radius is False


def test_min_max_radius(proposal):
    """Test configuration of min radius and no max radius"""
    FlowProposal.configure_min_max_radius(proposal, 5, 10)
    assert proposal.min_radius == 5.0
    assert proposal.max_radius == 10.0


@pytest.mark.parametrize('rmin, rmax', [(None, 1.0), (1.0, '2')])
def test_min_max_radius_invalid_input(proposal, rmin, rmax):
    """Test configuration of min radius and no max radius"""
    with pytest.raises(RuntimeError):
        FlowProposal.configure_min_max_radius(proposal, rmin, rmax)


@pytest.mark.parametrize('plot, plot_pool, plot_train',
                         [(True, True, True),
                          ('all', 'all', 'all'),
                          ('train', False, 'all'),
                          ('pool', 'all', False),
                          ('min', True, True),
                          ('minimal', True, True),
                          (False, False, False),
                          ('some', False, False)
                          ])
def test_configure_plotting(proposal, plot, plot_pool, plot_train):
    """Test the configuration of plotting settings"""
    FlowProposal.configure_plotting(proposal, plot)
    assert proposal._plot_pool == plot_pool
    assert proposal._plot_training == plot_train


@pytest.mark.parametrize('latent_prior, prior_func',
                         [('gaussian', 'draw_gaussian'),
                          ('truncated_gaussian', 'draw_truncated_gaussian'),
                          ('uniform', 'draw_uniform'),
                          ('uniform_nsphere', 'draw_nsphere'),
                          ('uniform_nball', 'draw_nsphere')
                          ])
def test_configure_latent_prior(proposal, latent_prior, prior_func):
    """Test to make sure the correct latent priors are used."""
    proposal.latent_prior = latent_prior
    proposal.flow_config = {'model_config': {}}
    FlowProposal.configure_latent_prior(proposal)
    assert proposal.draw_latent_prior == getattr(utils, prior_func)


def test_configure_latent_prior_var(proposal):
    """
    Test the latent prior when using a truncated Gaussian with a variance.
    """
    proposal.latent_prior = 'truncated_gaussian'
    proposal.flow_config = {'model_config': {'kwargs': {'var': 4}}}
    proposal.draw_latent_kwargs = {}
    FlowProposal.configure_latent_prior(proposal)
    assert proposal.draw_latent_prior == \
        getattr(utils, 'draw_truncated_gaussian')
    assert proposal.draw_latent_kwargs.get('var') == 4


def test_configure_latent_prior_unknown(proposal):
    """Make sure unknown latent priors raise an error"""
    proposal.latent_prior = 'truncated'
    with pytest.raises(RuntimeError) as excinfo:
        FlowProposal.configure_latent_prior(proposal)
    assert 'Unknown latent prior: truncated, ' in str(excinfo.value)


@pytest.mark.parametrize('inversion, parameters',
                         [(True, ['x', 'y']), (False, []), (['x'], ['x'])])
def test_set_boundary_inversion(proposal, inversion, parameters):
    """Check the correct parameters are inverted"""
    proposal.names = ['x', 'y', 'z']
    proposal.rescale_parameters = ['x', 'y']
    proposal.boundary_inversion = inversion
    proposal.inversion_type = 'split'
    FlowProposal.set_boundary_inversion(proposal)
    assert proposal.boundary_inversion == parameters


@pytest.mark.parametrize('rescale_parameters, parameters',
                         [(True, ['x', 'y']), (['x'], ['x'])])
def test_set_boundary_inversion_rescale_parameters(proposal,
                                                   rescale_parameters,
                                                   parameters):
    """Check the correct parameters depending on the type of rescale
    parameters.
    """
    proposal.names = ['x', 'y']
    proposal.rescale_parameters = rescale_parameters
    proposal.boundary_inversion = True
    proposal.inversion_type = 'split'
    FlowProposal.set_boundary_inversion(proposal)
    assert proposal.boundary_inversion == parameters


def test_set_boundary_inversion_no_rescaling(proposal):
    """Check boundary inversion is disabled if paraemters are not rescaled"""
    proposal.rescale_parameters = False
    proposal.boundary_inversion = ['x']
    with pytest.raises(RuntimeError) as excinfo:
        FlowProposal.set_boundary_inversion(proposal)
    assert 'Boundary inversion requires rescaling' in str(excinfo.value)


def test_set_boundary_inversion_incorrect_parameters(proposal):
    """Check an error is raised if a parameter is not recognised"""
    proposal.rescale_parameters = True
    proposal.boundary_inversion = ['x', 'z']
    proposal.names = ['x', 'y']
    with pytest.raises(RuntimeError) as excinfo:
        FlowProposal.set_boundary_inversion(proposal)
    assert 'Boundaries are not in known parameters' in str(excinfo.value)


def test_set_boundary_inversion_parameter_not_rescaled(proposal):
    """Check an error is raised if the parameter for inversion is not included
    in the rescaled parameters.
    """
    proposal.names = ['x', 'y']
    proposal.rescale_parameters = ['y']
    proposal.boundary_inversion = ['x']
    with pytest.raises(RuntimeError) as excinfo:
        FlowProposal.set_boundary_inversion(proposal)
    assert 'Boundaries are not in rescaled parameters' in str(excinfo.value)


def test_set_boundary_inversion_incorrect_inversion_type(proposal):
    """Check an error is raised if the inversion type is unknown"""
    proposal.names = ['x']
    proposal.rescale_parameters = True
    proposal.boundary_inversion = True
    proposal.inversion_type = 'half'
    with pytest.raises(RuntimeError) as excinfo:
        FlowProposal.set_boundary_inversion(proposal)
    assert 'Unknown inversion type' in str(excinfo.value)



# -*- coding: utf-8 -*-
"""Test different properties in FlowProposal"""
from nessai.proposal import FlowProposal


def test_poolsize(proposal):
    """Test poolsize property"""
    proposal._poolsize = 10
    proposal._poolsize_scale = 2
    assert FlowProposal.poolsize.__get__(proposal) == 20


def test_dims(proposal):
    """Test dims property"""
    proposal.names = ['x', 'y']
    assert FlowProposal.dims.__get__(proposal) == 2


def test_rescaled_dims(proposal):
    """Test rescaled_dims property"""
    proposal.rescaled_names = ['x', 'y']
    assert FlowProposal.rescaled_dims.__get__(proposal) == 2


def test_dtype(proposal):
    """Test dims property"""
    proposal.names = ['x', 'y']
    proposal._x_dtype = None
    assert FlowProposal.x_dtype.__get__(proposal) == \
        [('x', 'f8'), ('y', 'f8'), ('logP', 'f8'), ('logL', 'f8')]


def test_prime_dtype(proposal):
    """Test dims property"""
    proposal.rescaled_names = ['x', 'y']
    proposal._x_prime_dtype = None
    assert FlowProposal.x_prime_dtype.__get__(proposal) == \
        [('x', 'f8'), ('y', 'f8'), ('logP', 'f8'), ('logL', 'f8')]


def test_population_dtype(proposal):
    """Test dims property"""
    proposal.x_dtype = \
        [('x', 'f8'), ('y', 'f8'), ('logP', 'f8'), ('logL', 'f8')]
    proposal.use_x_prime_prior = False
    assert FlowProposal.population_dtype.__get__(proposal) == \
        [('x', 'f8'), ('y', 'f8'), ('logP', 'f8'), ('logL', 'f8')]


def test_population_dtype_prime_prior(proposal):
    """Test dims property"""
    proposal.x_prime_dtype = \
        [('x_p', 'f8'), ('y_p', 'f8'), ('logP', 'f8'), ('logL', 'f8')]
    proposal.use_x_prime_prior = True
    assert FlowProposal.population_dtype.__get__(proposal) == \
        [('x_p', 'f8'), ('y_p', 'f8'), ('logP', 'f8'), ('logL', 'f8')]



# -*- coding: utf-8 -*-
"""
Test the RescaleToBound class.
"""
import numpy as np
import pytest
from unittest.mock import create_autospec

from nessai.reparameterisations import RescaleToBounds
from nessai.livepoint import get_dtype, numpy_array_to_live_points


@pytest.fixture
def reparam():
    return create_autospec(RescaleToBounds)


@pytest.fixture()
def reparameterisation(model):
    def _get_reparameterisation(kwargs):
        return RescaleToBounds(parameters=model.names,
                               prior_bounds=model.bounds,
                               **kwargs)
    return _get_reparameterisation


@pytest.fixture(scope='function')
def assert_invertibility(model, n=100):
    def test_invertibility(reparam):
        x = model.new_point(N=n)
        x_prime = np.zeros([n], dtype=get_dtype(reparam.prime_parameters))
        log_j = 0

        assert x.size == x_prime.size

        x_re, x_prime_re, log_j_re = reparam.reparameterise(
            x, x_prime, log_j)

        np.testing.assert_array_equal(x, x_re)

        x_in = np.zeros([n], dtype=get_dtype(reparam.parameters))

        x_inv, x_prime_inv, log_j_inv = \
            reparam.inverse_reparameterise(x_in, x_prime_re, log_j)

        np.testing.assert_array_equal(x, x_inv)
        np.testing.assert_array_equal(x_prime_re, x_prime_inv)
        np.testing.assert_array_equal(log_j_re, -log_j_inv)

        return True

    return test_invertibility


@pytest.mark.parametrize('rescale_bounds', [None, [0, 1]])
def test_rescale_bounds(reparameterisation, assert_invertibility,
                        rescale_bounds):
    """Test the different options for rescale to bounds"""
    reparam = reparameterisation({'rescale_bounds': rescale_bounds})
    if rescale_bounds is None:
        rescale_bounds = {p: [-1, 1] for p in reparam.parameters}
    elif isinstance(rescale_bounds, list):
        rescale_bounds = {p: rescale_bounds for p in reparam.parameters}

    assert reparam.rescale_bounds == rescale_bounds
    assert assert_invertibility(reparam)


@pytest.mark.parametrize('boundary_inversion', [False,
                                                True,
                                                ['x']])
def test_boundary_inversion(reparameterisation, assert_invertibility,
                            boundary_inversion):
    """Test the different options for rescale to bounds"""
    reparam = reparameterisation({'boundary_inversion': boundary_inversion})

    assert assert_invertibility(reparam)


def test_set_bounds(reparam):
    """Test the set bounds method."""
    reparam.parameters = ['x']
    reparam.rescale_bounds = {'x': np.array([-1, 1])}
    reparam.pre_rescaling = lambda x: (x / 2, np.zeros_like(x))
    reparam.offsets = {'x': 1}
    RescaleToBounds.set_bounds(reparam, {'x': np.array([-10, 10])})
    np.testing.assert_array_equal(reparam.pre_prior_bounds['x'], [-5, 5])
    np.testing.assert_array_equal(reparam.bounds['x'], [-6, 4])


@pytest.mark.integration_test
def test_update_prime_prior_bounds_integration():
    """Assert the prime prior bounds are correctly computed"""
    rescaling = (
        lambda x: (x / 2, np.zeros_like(x)),
        lambda x: (2 * x, np.zeros_like(x)),
    )
    reparam = RescaleToBounds(
        parameters=['x'], prior_bounds=[1000, 1001], prior='uniform',
        pre_rescaling=rescaling, offset=True,
    )
    np.testing.assert_equal(reparam.offsets['x'], 500.25)
    np.testing.assert_array_equal(reparam.prior_bounds['x'], [1000, 1001])
    np.testing.assert_array_equal(reparam.pre_prior_bounds['x'], [500, 500.5])
    np.testing.assert_array_equal(reparam.bounds['x'], [-0.25, 0.25])
    np.testing.assert_array_equal(
        reparam.prime_prior_bounds['x_prime'], [-1, 1]
    )

    x_prime = numpy_array_to_live_points(
        np.array([[-2], [-1], [0.5], [1], [10]]), ['x_prime']
    )
    log_prior = reparam.x_prime_log_prior(x_prime)
    expected = np.array([-np.inf, 0, 0, 0, -np.inf])
    np.testing.assert_equal(log_prior, expected)
